"""
Automated Market Making liquidity pair.
https://docs.uniswap.org/protocol/V2/reference/smart-contracts/pair
"""

from web3 import Web3
from .utils.utils import swap_expected_amount1

ABI = '''
    [
        {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"owner","type":"address"},{"indexed":true,"internalType":"address","name":"spender","type":"address"},{"indexed":false,"internalType":"uint256","name":"value","type":"uint256"}],"name":"Approval","type":"event"},
        {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"sender","type":"address"},{"indexed":false,"internalType":"uint256","name":"amount0","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"amount1","type":"uint256"},{"indexed":true,"internalType":"address","name":"to","type":"address"}],"name":"Burn","type":"event"},
        {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"sender","type":"address"},{"indexed":false,"internalType":"uint256","name":"amount0","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"amount1","type":"uint256"}],"name":"Mint","type":"event"},
        {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"sender","type":"address"},{"indexed":false,"internalType":"uint256","name":"amount0In","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"amount1In","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"amount0Out","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"amount1Out","type":"uint256"},{"indexed":true,"internalType":"address","name":"to","type":"address"}],"name":"Swap","type":"event"},
        {"anonymous":false,"inputs":[{"indexed":false,"internalType":"uint112","name":"reserve0","type":"uint112"},{"indexed":false,"internalType":"uint112","name":"reserve1","type":"uint112"}],"name":"Sync","type":"event"},
        {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"from","type":"address"},{"indexed":true,"internalType":"address","name":"to","type":"address"},{"indexed":false,"internalType":"uint256","name":"value","type":"uint256"}],"name":"Transfer","type":"event"},
        {"inputs":[],"name":"DOMAIN_SEPARATOR","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
        {"inputs":[],"name":"MINIMUM_LIQUIDITY","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"pure","type":"function"},
        {"inputs":[],"name":"PERMIT_TYPEHASH","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"pure","type":"function"},
        {"inputs":[{"internalType":"address","name":"owner","type":"address"},{"internalType":"address","name":"spender","type":"address"}],"name":"allowance","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
        {"inputs":[{"internalType":"address","name":"spender","type":"address"},{"internalType":"uint256","name":"value","type":"uint256"}],"name":"approve","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"address","name":"owner","type":"address"}],"name":"balanceOf","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
        {"inputs":[{"internalType":"address","name":"to","type":"address"}],"name":"burn","outputs":[{"internalType":"uint256","name":"amount0","type":"uint256"},{"internalType":"uint256","name":"amount1","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[],"name":"decimals","outputs":[{"internalType":"uint8","name":"","type":"uint8"}],"stateMutability":"pure","type":"function"},
        {"inputs":[],"name":"factory","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},
        {"inputs":[],"name":"getReserves","outputs":[{"internalType":"uint112","name":"reserve0","type":"uint112"},{"internalType":"uint112","name":"reserve1","type":"uint112"},{"internalType":"uint32","name":"blockTimestampLast","type":"uint32"}],"stateMutability":"view","type":"function"},
        {"inputs":[{"internalType":"address","name":"","type":"address"},{"internalType":"address","name":"","type":"address"}],"name":"initialize","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[],"name":"kLast","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
        {"inputs":[{"internalType":"address","name":"to","type":"address"}],"name":"mint","outputs":[{"internalType":"uint256","name":"liquidity","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[],"name":"name","outputs":[{"internalType":"string","name":"","type":"string"}],"stateMutability":"pure","type":"function"},
        {"inputs":[{"internalType":"address","name":"owner","type":"address"}],"name":"nonces","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
        {"inputs":[{"internalType":"address","name":"owner","type":"address"},{"internalType":"address","name":"spender","type":"address"},{"internalType":"uint256","name":"value","type":"uint256"},{"internalType":"uint256","name":"deadline","type":"uint256"},{"internalType":"uint8","name":"v","type":"uint8"},{"internalType":"bytes32","name":"r","type":"bytes32"},{"internalType":"bytes32","name":"s","type":"bytes32"}],"name":"permit","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[],"name":"price0CumulativeLast","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
        {"inputs":[],"name":"price1CumulativeLast","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
        {"inputs":[{"internalType":"address","name":"to","type":"address"}],"name":"skim","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"uint256","name":"amount0Out","type":"uint256"},{"internalType":"uint256","name":"amount1Out","type":"uint256"},{"internalType":"address","name":"to","type":"address"},{"internalType":"bytes","name":"data","type":"bytes"}],"name":"swap","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[],"name":"symbol","outputs":[{"internalType":"string","name":"","type":"string"}],"stateMutability":"pure","type":"function"},
        {"inputs":[],"name":"sync","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[],"name":"token0","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},
        {"inputs":[],"name":"token1","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},
        {"inputs":[],"name":"totalSupply","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
        {"inputs":[{"internalType":"address","name":"to","type":"address"},{"internalType":"uint256","name":"value","type":"uint256"}],"name":"transfer","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"address","name":"from","type":"address"},{"internalType":"address","name":"to","type":"address"},{"internalType":"uint256","name":"value","type":"uint256"}],"name":"transferFrom","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"nonpayable","type":"function"}
    ]
    '''


def block_explorer_link(txid):
    return 'https://explorer.harmony.one/tx/' + str(txid)


def swap(pool_address, amount0_out, amount1_out, to, private_key, nonce, gas_price_gwei, tx_timeout_seconds, rpc_address, logger):
    w3 = Web3(Web3.HTTPProvider(rpc_address))
    account = w3.eth.account.privateKeyToAccount(private_key)
    w3.eth.default_account = account.address

    contract_address = Web3.toChecksumAddress(pool_address)
    contract = w3.eth.contract(contract_address, abi=ABI)

    tx = contract.functions.startQuestWithData(amount0_out, amount1_out, to, None)

    if isinstance(gas_price_gwei, dict):  # dynamic fee
        tx = tx.buildTransaction(
            {'maxFeePerGas': w3.toWei(gas_price_gwei['maxFeePerGas'], 'gwei'),
             'maxPriorityFeePerGas': w3.toWei(gas_price_gwei['maxPriorityFeePerGas'], 'gwei'), 'nonce': nonce})
    else:  # legacy
        tx = tx.buildTransaction({'gasPrice': w3.toWei(gas_price_gwei, 'gwei'), 'nonce': nonce})

    logger.debug("Signing transaction")
    signed_tx = w3.eth.account.sign_transaction(tx, private_key=private_key)
    logger.debug("Sending transaction " + str(tx))
    ret = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    logger.debug("Transaction successfully sent !")
    logger.info(
        "Waiting for transaction " + block_explorer_link(signed_tx.hash.hex()) + " to be mined")

    tx_receipt = w3.eth.wait_for_transaction_receipt(transaction_hash=signed_tx.hash, timeout=tx_timeout_seconds,
                                                     poll_latency=2)
    logger.info("Transaction mined !")

    return tx_receipt


def name(pool_address, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract_address = Web3.toChecksumAddress(pool_address)
    contract = w3.eth.contract(contract_address, abi=ABI)

    return contract.functions.symbol().call()


def symbol(pool_address, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract_address = Web3.toChecksumAddress(pool_address)
    contract = w3.eth.contract(contract_address, abi=ABI)

    return contract.functions.symbol().call()


def token_0(pool_address, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract_address = Web3.toChecksumAddress(pool_address)
    contract = w3.eth.contract(contract_address, abi=ABI)

    return contract.functions.token0().call()


def token_1(pool_address, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract_address = Web3.toChecksumAddress(pool_address)
    contract = w3.eth.contract(contract_address, abi=ABI)

    return contract.functions.token1().call()


def decimals(pool_address, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract_address = Web3.toChecksumAddress(pool_address)
    contract = w3.eth.contract(contract_address, abi=ABI)

    return contract.functions.decimals().call()


def total_supply(pool_address, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract_address = Web3.toChecksumAddress(pool_address)
    contract = w3.eth.contract(contract_address, abi=ABI)

    return contract.functions.totalSupply().call()


def get_reserves(pool_address, rpc_address):
    '''
    Returns the reserves of token0 and token1 used to price trades and distribute liquidity.
    Also returns the block.timestamp (mod 2**32) of the last block during which an interaction occurred for the pair.
    :param pool_address:
    :param rpc_address:
    :return: reserve0, reserve1, blockTimestampLast
    '''
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract_address = Web3.toChecksumAddress(pool_address)
    contract = w3.eth.contract(contract_address, abi=ABI)

    return contract.functions.getReserves().call()


def balance_of(pool_address, owner_address, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract_address = Web3.toChecksumAddress(pool_address)
    contract = w3.eth.contract(contract_address, abi=ABI)

    return contract.functions.balanceOf(owner_address).call()


def price_0_cumulative_last(pool_address, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract_address = Web3.toChecksumAddress(pool_address)
    contract = w3.eth.contract(contract_address, abi=ABI)

    return contract.functions.price0CumulativeLast().call()


def price_1_cumulative_last(pool_address, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract_address = Web3.toChecksumAddress(pool_address)
    contract = w3.eth.contract(contract_address, abi=ABI)

    return contract.functions.price1CumulativeLast().call()


class UniswapV2Pair:
    def __init__(self, pair_address, rpc_address, logger):
        self.address = pair_address
        self.rpc_address = rpc_address
        self.logger = logger

        self.symbol_value = None
        self.token_0_value = None
        self.token_1_value = None

    def symbol(self):
        if self.symbol_value is None:
            self.symbol_value = symbol(self.address, self.rpc_address)
        return self.symbol_value

    def token_0(self):
        if self.token_0_value is None:
            self.token_0_value = token_0(self.address, self.rpc_address)
        return self.token_0_value

    def token_1(self):
        if self.token_1_value is None:
            self.token_1_value = token_1(self.address, self.rpc_address)
        return self.token_1_value

    def decimals(self):
        return decimals(self.address, self.rpc_address)

    def total_supply(self):
        return total_supply(self.address, self.rpc_address)

    def reserves(self):
        return get_reserves(self.address, self.rpc_address)

    def balance_of(self, address):
        return balance_of(self.address, address, self.rpc_address)

    def price_0_cumulative_last(self):
        return price_0_cumulative_last(self.address, self.rpc_address)

    def price_1_cumulative_last(self):
        return price_1_cumulative_last(self.address, self.rpc_address)

    def expected_amount1(self, amount0):
        reserves = get_reserves(self.address, self.rpc_address)
        return swap_expected_amount1(reserves[0], reserves[1], amount0)

    def expected_amount0(self, amount1):
        reserves = get_reserves(self.address, self.rpc_address)
        return swap_expected_amount1(reserves[1], reserves[0], amount1)


import dex.erc20 as tokens
import logging
import sys
from web3 import Web3


if __name__ == "__main__":
    log_format = '%(asctime)s|%(name)s|%(levelname)s: %(message)s'

    logger = logging.getLogger("DFK-erc20")
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.INFO, format=log_format, stream=sys.stdout)

    rpc_server = 'https://api.harmony.one'
    logger.info("Using RPC server " + rpc_server)

    w3 = Web3(Web3.HTTPProvider(rpc_server))

    realm = 'serendale'
    serendale_items = tokens.get_realm_item_list(realm)
    logger.info('{} items in {}'.format(len(serendale_items), realm))
    token_address = tokens.symbol2address('JEWEL', realm)
    name = tokens.name(token_address, rpc_server)
    symbol = tokens.symbol(token_address, rpc_server)
    decimal = tokens.decimals(token_address, rpc_server)
    balance = tokens.balance_of('0x2E7669F61eA77F02445A015FBdcFe2DE47083E02', token_address, rpc_server)
    balance = tokens.wei2eth(w3, balance)
    logger.info(name + " -> " + str(balance) + " " + symbol)




# stone_carver.py

from web3 import Web3

SERENDALE_CONTRACT_ADDRESS = "0xfFB8a55676edA75954AB45a6Ce16F88b119dC511"
CRYSTALVALE_CONTRACT_ADDRESS = None

ABI = '''
	[
		{"anonymous":false,"inputs":[{"indexed":false,"internalType":"uint8","name":"version","type":"uint8"}],"name":"Initialized","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"account","type":"address"}],"name":"Paused","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"stoneAddress","type":"address"},{"indexed":false,"internalType":"address[]","name":"requiredResources","type":"address[]"},{"indexed":false,"internalType":"uint32[]","name":"requiredQuantities","type":"uint32[]"},{"indexed":false,"internalType":"bool","name":"active","type":"bool"}],"name":"RecipeSet","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"previousAdminRole","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"newAdminRole","type":"bytes32"}],"name":"RoleAdminChanged","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleGranted","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleRevoked","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":false,"internalType":"uint256","name":"workingUntil","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"awayUntil","type":"uint256"}],"name":"ShopSetUp","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"player","type":"address"},{"indexed":false,"internalType":"address","name":"stoneAddress","type":"address"},{"indexed":false,"internalType":"uint256","name":"quantity","type":"uint256"}],"name":"StoneCarved","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"account","type":"address"}],"name":"Unpaused","type":"event"},
		{"inputs":[],"name":"DEFAULT_ADMIN_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
		{"inputs":[],"name":"MODERATOR_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
		{"inputs":[],"name":"adminSetupShop","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[],"name":"awayUntil","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"address","name":"_stoneAddress","type":"address"},{"internalType":"uint256","name":"_quantity","type":"uint256"}],"name":"carveStone","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[],"name":"getAvailability","outputs":[{"internalType":"uint256","name":"","type":"uint256"},{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"address","name":"stoneAddress","type":"address"}],"name":"getRecipe","outputs":[{"components":[{"internalType":"address[]","name":"requiredResources","type":"address[]"},{"internalType":"uint32[]","name":"requiredQuantities","type":"uint32[]"},{"internalType":"bool","name":"active","type":"bool"}],"internalType":"struct StoneCarver.Recipe","name":"","type":"tuple"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"}],"name":"getRoleAdmin","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"grantRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"hasRole","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
		{"inputs":[],"name":"initialize","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[],"name":"minClosedTime","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
		{"inputs":[],"name":"minOpenTime","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
		{"inputs":[],"name":"pause","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[],"name":"paused","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"address","name":"","type":"address"}],"name":"recipes","outputs":[{"internalType":"bool","name":"active","type":"bool"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"renounceRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"revokeRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[{"internalType":"address","name":"_stoneAddress","type":"address"},{"internalType":"address[]","name":"_requiredResources","type":"address[]"},{"internalType":"uint32[]","name":"_requiredQuantities","type":"uint32[]"},{"internalType":"bool","name":"_active","type":"bool"}],"name":"setRecipe","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[{"internalType":"uint256","name":"_minOpenTime","type":"uint256"},{"internalType":"uint256","name":"_varOpenTime","type":"uint256"},{"internalType":"uint256","name":"_minClosedTime","type":"uint256"},{"internalType":"uint256","name":"_varClosedTime","type":"uint256"}],"name":"setTimes","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[],"name":"setUpShop","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[{"internalType":"bytes4","name":"interfaceId","type":"bytes4"}],"name":"supportsInterface","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
		{"inputs":[],"name":"unpause","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[],"name":"varClosedTime","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
		{"inputs":[],"name":"varOpenTime","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
		{"inputs":[],"name":"workingUntil","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"}
	]
'''


def block_explorer_link(contract_address, txid):
	if hasattr(contract_address, 'address'):
		contract_address = str(contract_address.address)
	contract_address = str(contract_address).upper()
	if contract_address == SERENDALE_CONTRACT_ADDRESS.upper():
		return 'https://explorer.harmony.one/tx/' + str(txid)
	elif contract_address == CRYSTALVALE_CONTRACT_ADDRESS.upper():
		return 'https://subnets.avax.network/defi-kingdoms/dfk-chain/explorer/tx/' + str(txid)
	else:
		return str(txid)


def away_until(contract_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	contract = w3.eth.contract(Web3.toChecksumAddress(contract_address), abi=ABI)
	return contract.functions.awayUntil().call()


def carve_stone(contract_address, item_address, quantity, private_key, nonce, gas_price_gwei, tx_timeout_seconds,
				rpc_address, logger):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	account = w3.eth.account.privateKeyToAccount(private_key)
	w3.eth.default_account = account.address

	contract_address = Web3.toChecksumAddress(contract_address)
	contract = w3.eth.contract(contract_address, abi=ABI)

	if isinstance(gas_price_gwei, dict):  # dynamic fee
		tx = contract.functions.carveStone(item_address, quantity).buildTransaction(
			{'maxFeePerGas': w3.toWei(gas_price_gwei['maxFeePerGas'], 'gwei'),
			 'maxPriorityFeePerGas': w3.toWei(gas_price_gwei['maxPriorityFeePerGas'], 'gwei'), 'nonce': nonce})
	else:  # legacy
		tx = contract.functions.carveStone(item_address, quantity).buildTransaction(
			{'gasPrice': w3.toWei(gas_price_gwei, 'gwei'), 'nonce': nonce})

	logger.debug("Signing transaction")
	signed_tx = w3.eth.account.sign_transaction(tx, private_key=private_key)
	logger.debug("Sending transaction " + str(tx))
	ret = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
	logger.debug("Transaction successfully sent !")
	logger.info(
		"Waiting for transaction " + block_explorer_link(contract_address, signed_tx.hash.hex()) + " to be mined")
	tx_receipt = w3.eth.wait_for_transaction_receipt(transaction_hash=signed_tx.hash, timeout=tx_timeout_seconds,
													 poll_latency=2)
	logger.info("Transaction mined !")
	return tx_receipt


def get_availability(contract_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	contract = w3.eth.contract(Web3.toChecksumAddress(contract_address), abi=ABI)
	return contract.functions.getAvailability().call()


def get_recipe(item_address, contract_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	contract = w3.eth.contract(Web3.toChecksumAddress(contract_address), abi=ABI)
	return contract.functions.getRecipe(item_address).call()


def min_closed_time(contract_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	contract = w3.eth.contract(Web3.toChecksumAddress(contract_address), abi=ABI)
	return contract.functions.minClosedTime().call()


def min_open_time(contract_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	contract = w3.eth.contract(Web3.toChecksumAddress(contract_address), abi=ABI)
	return contract.functions.minOpenTime().call()


def recipes(item_address, contract_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	contract = w3.eth.contract(Web3.toChecksumAddress(contract_address), abi=ABI)
	return contract.functions.recipes(item_address).call()


def set_up_shop(contract_address, private_key, nonce, gas_price_gwei, tx_timeout_seconds, rpc_address, logger):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	account = w3.eth.account.privateKeyToAccount(private_key)
	w3.eth.default_account = account.address

	contract_address = Web3.toChecksumAddress(contract_address)
	contract = w3.eth.contract(contract_address, abi=ABI)

	if isinstance(gas_price_gwei, dict):  # dynamic fee
		tx = contract.functions.setUpShop().buildTransaction(
			{'maxFeePerGas': w3.toWei(gas_price_gwei['maxFeePerGas'], 'gwei'),
			 'maxPriorityFeePerGas': w3.toWei(gas_price_gwei['maxPriorityFeePerGas'], 'gwei'), 'nonce': nonce})
	else:  # legacy
		tx = contract.functions.setUpShop().buildTransaction(
			{'gasPrice': w3.toWei(gas_price_gwei, 'gwei'), 'nonce': nonce})

	logger.debug("Signing transaction")
	signed_tx = w3.eth.account.sign_transaction(tx, private_key=private_key)
	logger.debug("Sending transaction " + str(tx))
	ret = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
	logger.debug("Transaction successfully sent !")
	logger.info(
		"Waiting for transaction " + block_explorer_link(contract_address, signed_tx.hash.hex()) + " to be mined")
	tx_receipt = w3.eth.wait_for_transaction_receipt(transaction_hash=signed_tx.hash, timeout=tx_timeout_seconds,
													 poll_latency=2)
	logger.info("Transaction mined !")
	return tx_receipt


def var_closed_time(contract_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	contract = w3.eth.contract(Web3.toChecksumAddress(contract_address), abi=ABI)
	return contract.functions.varClosedTime().call()


def var_open_time(contract_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	contract = w3.eth.contract(Web3.toChecksumAddress(contract_address), abi=ABI)
	return contract.functions.varOpenTime().call()


def working_until(contract_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	contract = w3.eth.contract(Web3.toChecksumAddress(contract_address), abi=ABI)
	return contract.functions.workingUntil().call()




def swap_expected_amount1(reserve0, reserve1, amount0_wei=1):
    p = reserve0 / reserve1
    amount1_wei = amount0_wei / p
    p2 = (reserve0 + amount0_wei) / (reserve1 - amount1_wei)
    return (amount1_wei + amount0_wei / p2) / 2


def human_readable_pool_info(pool_info):
    if pool_info is None:
        return None

    human_readable = {}
    human_readable['address'] = pool_info[0]
    human_readable['allocPoint'] = pool_info[1]
    human_readable['lastRewardBlock'] = pool_info[2]
    human_readable['accGovTokenPerShare'] = pool_info[3]

    return human_readable


def human_readable_user_info(user_info):
    if user_info is None:
        return None

    human_readable = {}
    human_readable['amount'] = user_info[0]
    human_readable['rewardDebt'] = user_info[1]
    human_readable['rewardDebtAtBlock'] = user_info[2]
    human_readable['lastWithdrawBlock'] = user_info[3]
    human_readable['firstDepositBlock'] = user_info[4]
    human_readable['blockdelta'] = user_info[5]
    human_readable['lastDepositBlock'] = user_info[6]

    return human_readable



from web3 import Web3

SERENDALE_CONTRACT_ADDRESS = "0x87cba8f998f902f2fff990effa1e261f35932e57"
CRYSTALVALE_CONTRACT_ADDRESS = "0x2542e1Ce063FED3b5Aa81936c5a8f6Eeccaa6B4A"

ABI = '''
	[
		{"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"account","type":"address"}],"name":"Paused","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"potionAddress","type":"address"},{"indexed":false,"internalType":"address[]","name":"requiredResources","type":"address[]"},{"indexed":false,"internalType":"uint32[]","name":"requiredQuantities","type":"uint32[]"}],"name":"PotionAdded","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"player","type":"address"},{"indexed":false,"internalType":"address","name":"potionAddress","type":"address"},{"indexed":false,"internalType":"uint256","name":"quantity","type":"uint256"},{"indexed":false,"internalType":"address[]","name":"requiredResources","type":"address[]"},{"indexed":false,"internalType":"uint32[]","name":"requiredQuantities","type":"uint32[]"}],"name":"PotionCreated","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"potionAddress","type":"address"},{"indexed":false,"internalType":"address[]","name":"requiredResources","type":"address[]"},{"indexed":false,"internalType":"uint32[]","name":"requiredQuantities","type":"uint32[]"},{"indexed":false,"internalType":"uint8","name":"status","type":"uint8"}],"name":"PotionUpdated","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"previousAdminRole","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"newAdminRole","type":"bytes32"}],"name":"RoleAdminChanged","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleGranted","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleRevoked","type":"event"},
		{"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"account","type":"address"}],"name":"Unpaused","type":"event"},
		{"inputs":[],"name":"DEFAULT_ADMIN_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
		{"inputs":[],"name":"MODERATOR_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"address","name":"_potionAddress","type":"address"},{"internalType":"address[]","name":"_requiredResources","type":"address[]"},{"internalType":"uint32[]","name":"_requiredQuantities","type":"uint32[]"}],"name":"addPotion","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[{"internalType":"address","name":"","type":"address"}],"name":"addressToPotionId","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"address","name":"_potionAddress","type":"address"},{"internalType":"uint256","name":"_quantity","type":"uint256"}],"name":"createPotion","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[{"internalType":"address","name":"_potionAddress","type":"address"}],"name":"getPotion","outputs":[{"components":[{"internalType":"address","name":"potionAddress","type":"address"},{"internalType":"address[]","name":"requiredResources","type":"address[]"},{"internalType":"uint32[]","name":"requiredQuantities","type":"uint32[]"},{"internalType":"uint8","name":"status","type":"uint8"}],"internalType":"struct Alchemist.Potion","name":"","type":"tuple"}],"stateMutability":"view","type":"function"},
		{"inputs":[],"name":"getPotions","outputs":[{"components":[{"internalType":"address","name":"potionAddress","type":"address"},{"internalType":"address[]","name":"requiredResources","type":"address[]"},{"internalType":"uint32[]","name":"requiredQuantities","type":"uint32[]"},{"internalType":"uint8","name":"status","type":"uint8"}],"internalType":"struct Alchemist.Potion[]","name":"","type":"tuple[]"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"}],"name":"getRoleAdmin","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"grantRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"hasRole","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"address","name":"_dfkGoldAddress","type":"address"}],"name":"initialize","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[],"name":"paused","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"uint256","name":"","type":"uint256"}],"name":"potions","outputs":[{"internalType":"address","name":"potionAddress","type":"address"},{"internalType":"uint8","name":"status","type":"uint8"}],"stateMutability":"view","type":"function"},
		{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"renounceRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"revokeRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[{"internalType":"bytes4","name":"interfaceId","type":"bytes4"}],"name":"supportsInterface","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
		{"inputs":[],"name":"togglePause","outputs":[],"stateMutability":"nonpayable","type":"function"},
		{"inputs":[{"internalType":"address","name":"_potionAddress","type":"address"},{"internalType":"address[]","name":"_requiredResources","type":"address[]"},{"internalType":"uint32[]","name":"_requiredQuantities","type":"uint32[]"},{"internalType":"uint8","name":"_status","type":"uint8"}],"name":"updatePotion","outputs":[],"stateMutability":"nonpayable","type":"function"}
	]
'''


def block_explorer_link(contract_address, txid):
	if hasattr(contract_address, 'address'):
		contract_address = str(contract_address.address)
	contract_address = str(contract_address).upper()
	if contract_address == SERENDALE_CONTRACT_ADDRESS.upper():
		return 'https://explorer.harmony.one/tx/' + str(txid)
	elif contract_address == CRYSTALVALE_CONTRACT_ADDRESS.upper():
		return 'https://subnets.avax.network/defi-kingdoms/dfk-chain/explorer/tx/' + str(txid)
	else:
		return str(txid)


def create_potion(contract_address, potion_address, quantity, private_key, nonce, gas_price_gwei, tx_timeout_seconds, rpc_address, logger):
	w3 = Web3(Web3.HTTPProvider(rpc_address))
	account = w3.eth.account.privateKeyToAccount(private_key)
	w3.eth.default_account = account.address

	contract_address = Web3.toChecksumAddress(contract_address)
	contract = w3.eth.contract(contract_address, abi=ABI)

	if isinstance(gas_price_gwei, dict):  # dynamic fee
		tx = contract.functions.createPotion(potion_address, quantity).buildTransaction(
			{'maxFeePerGas': w3.toWei(gas_price_gwei['maxFeePerGas'], 'gwei'),
			 'maxPriorityFeePerGas': w3.toWei(gas_price_gwei['maxPriorityFeePerGas'], 'gwei'), 'nonce': nonce})
	else:  # legacy
		tx = contract.functions.createPotion(potion_address, quantity).buildTransaction(
			{'gasPrice': w3.toWei(gas_price_gwei, 'gwei'), 'nonce': nonce})

	logger.debug("Signing transaction")
	signed_tx = w3.eth.account.sign_transaction(tx, private_key=private_key)
	logger.debug("Sending transaction " + str(tx))
	ret = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
	logger.debug("Transaction successfully sent !")
	logger.info(
		"Waiting for transaction " + block_explorer_link(contract_address, signed_tx.hash.hex()) + " to be mined")

	tx_receipt = w3.eth.wait_for_transaction_receipt(transaction_hash=signed_tx.hash, timeout=tx_timeout_seconds,
													 poll_latency=2)
	logger.info("Transaction mined !")

	return tx_receipt


def address_to_potion_id(contract_address, potion_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))

	contract_address = Web3.toChecksumAddress(contract_address)
	contract = w3.eth.contract(contract_address, abi=ABI)

	return contract.functions.addressToPotionId(potion_address).call()


def potion_id_to_address_amount(contract_address, uint256, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))

	contract_address = Web3.toChecksumAddress(contract_address)
	contract = w3.eth.contract(contract_address, abi=ABI)

	raw = contract.functions.potions(uint256).call()
	return {'address': raw[0], 'batchSize': raw[1]}


def get_potion(contract_address, potion_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))

	contract_address = Web3.toChecksumAddress(contract_address)
	contract = w3.eth.contract(contract_address, abi=ABI)

	raw = contract.functions.getPotion(potion_address).call()
	return {'address': raw[0], 'ingredientAddresses': raw[1], 'ingredientQuantities': raw[2], 'batchSize': raw[3]}


def get_potions(contract_address, rpc_address):
	w3 = Web3(Web3.HTTPProvider(rpc_address))

	contract_address = Web3.toChecksumAddress(contract_address)
	contract = w3.eth.contract(contract_address, abi=ABI)

	raw = contract.functions.getPotions().call()
	potions = []
	for r in raw:
		potions.append({'address': r[0], 'ingredientAddresses': r[1], 'ingredientQuantities': r[2], 'batchSize': r[3]})

	return potions



def human_readable_raffle(raffle):
    human_readable = {}
    human_readable['id'] = raffle[0]
    human_readable['raffleType'] = raffle[1]
    human_readable['startTime'] = raffle[2]
    human_readable['totalEntries'] = raffle[3]
    human_readable['endTime'] = raffle[4]
    human_readable['closedBlock'] = raffle[5]
    human_readable['winners'] = raffle[6]
    human_readable['status'] = raffle[7]

    return human_readable


def human_readable_raffle_type(raffle):
    human_readable = {}
    human_readable['id'] = raffle[0]
    human_readable['rewards'] = raffle[1]
    human_readable['rewardAmounts'] = raffle[2]
    human_readable['maxWinners'] = raffle[3]
    human_readable['duration'] = raffle[4]

    return human_readable



import requests

AUCTIONS_OPEN_GRAPHQL_QUERY = """
                        query {
                          assistingAuctions(skip: %d, first: %d, orderBy: startedAt, orderDirection: desc, where: {open: true}) {
                            id
                            seller {
                                name
                            }
                            tokenId {
                              id
                              owner {
                                owner
                              }                        
                              statGenes
                              generation
                              rarity
                              mainClass
                              subClass
                              summons
                              maxSummons
                              summonerId {
                                id
                              }
                              assistantId {
                                id
                              }
                            }
                            startingPrice
                            endingPrice
                            startedAt
                            duration
                            winner {
                              id
                              name
                            }
                            open
                            
                          }
                          
                        }
                        """


def get_open_auctions(graphql_address, skip=0, count=1000):

    r = requests.post(graphql_address, json={'query': AUCTIONS_OPEN_GRAPHQL_QUERY % (skip, count)})
    if r.status_code != 200:
        raise Exception("HTTP error " + str(r.status_code) + ": " + r.text)
    data = r.json()
    return data['data']['assistingAuctions']




import json

FAIL_ON_NOT_FOUND = False

ALPHABET = '123456789abcdefghijkmnopqrstuvwx'

CRYSTALEVALE_HERO_OFFSET = 1_000_000_000_000

rarity = {
    0: "common",
    1: "uncommon",
    2: "rare",
    3: "legendary",
    4: "mythic",
}

_class = {
    0: "warrior",
    1: "knight",
    2: "thief",
    3: "archer",
    4: "priest",
    5: "wizard",
    6: "monk",
    7: "pirate",
    8: "berserker",
    9: "seer",
    16: "paladin",
    17: "darkknight",
    18: "summoner",
    19: "ninja",
    20: "shapeshifter",
    24: "dragoon",
    25: "sage",
    28: "dreadknight"
}

visual_traits = {
    0: 'gender',
    1: 'headAppendage',
    2: 'backAppendage',
    3: 'background',
    4: 'hairStyle',
    5: 'hairColor',
    6: 'visualUnknown1',
    7: 'eyeColor',
    8: 'skinColor',
    9: 'appendageColor',
    10: 'backAppendageColor',
    11: 'visualUnknown2'
}

stat_traits = {
    0: "class",
    1: "subClass",
    2: "profession",
    3: "passive1",
    4: "passive2",
    5: "active1",
    6: "active2",
    7: "statBoost1",
    8: "statBoost2",
    9: "statsUnknown1",
    10: "element",
    11: "statsUnknown2"
}

ability_traits = {
    0: "basic1",
    1: "basic2",
    2: "basic3",
    3: "basic4",
    4: "basic5",
    5: "basic6",
    6: "basic7",
    7: "basic8",
    16: "advanced1",
    17: "advanced2",
    18: "advanced3",
    19: "advanced4",
    24: "elite1",
    25: "elite2",
    28: "transcendant1"
}

professions = {
    0: 'mining',
    2: 'gardening',
    4: 'fishing',
    6: 'foraging',
}

stats = {
    0: 'strength',
    2: 'agility',
    4: 'intelligence',
    6: 'wisdom',
    8: 'luck',
    10: 'vitality',
    12: 'endurance',
    14: 'dexterity'
}

elements = {
    0: 'fire',
    2: 'water',
    4: 'earth',
    6: 'wind',
    8: 'lightning',
    10: 'ice',
    12: 'light',
    14: 'dark',
}


def cv2sd_cv_hero_id(cv_hero_id):
    return cv_hero_id - CRYSTALEVALE_HERO_OFFSET


def sd2cv_cv_hero_id(cv_hero_id):
    return cv_hero_id + CRYSTALEVALE_HERO_OFFSET


def parse_rarity(id):
    value = rarity.get(id, None)
    if FAIL_ON_NOT_FOUND and value is None:
        raise Exception("Rarity not found")
    return value


def parse_class(id):
    value = _class.get(id, None)
    if FAIL_ON_NOT_FOUND and value is None:
        raise Exception("Class not found")
    return value


def parse_profession(id):
    value = professions.get(id, None)
    if FAIL_ON_NOT_FOUND and value is None:
        raise Exception("Profession not found")
    return value


def parse_stat(id):
    value = stats.get(id, None)
    if FAIL_ON_NOT_FOUND and value is None:
        raise Exception("Stat not found")
    return value


def parse_ability(id):
    value = ability_traits.get(id, None)
    if FAIL_ON_NOT_FOUND and value is None:
        raise Exception("Ability not found")
    return value


def parse_element(id):
    value = elements.get(id, None)
    if FAIL_ON_NOT_FOUND and value is None:
        raise Exception("Element not found")
    return value


def genes2traits(genes):
    traits = []

    stat_raw_kai = "".join(__genesToKai(genes).split(' '))
    for ki in range(0, len(stat_raw_kai)):
        kai = stat_raw_kai[ki]
        value_num = __kai2dec(kai)
        traits.append(value_num)

    assert len(traits) == 48
    arranged_traits = [[], [], [], []]
    for i in range(0, 12):
        index = i << 2
        for j in range(0, len(arranged_traits)):
            arranged_traits[j].append(traits[index + j])

    arranged_traits.reverse()
    return arranged_traits


def parse_stat_genes(genes):
    traits = genes2traits(genes)
    stats = parse_stat_trait(traits[0])
    r1 = parse_stat_trait(traits[1])
    r2 = parse_stat_trait(traits[2])
    r3 = parse_stat_trait(traits[3])

    stats['r1'] = r1
    stats['r2'] = r2
    stats['r3'] = r3
    stats['raw'] = genes

    return stats


def parse_stat_trait(trait):

    if len(trait) != 12:
        raise Exception("Traits must be an array of 12")

    stats = {}
    for i in range(0, 12):
        stat_trait = stat_traits.get(i, None)
        stats[stat_trait] = trait[i]

    stats['class'] = parse_class(stats['class'])
    stats['subClass'] = parse_class(stats['subClass'])

    stats['profession'] = parse_profession(stats['profession'])

    stats['passive1'] = parse_ability(stats['passive1'])
    stats['passive2'] = parse_ability(stats['passive2'])
    stats['active1'] = parse_ability(stats['active1'])
    stats['active2'] = parse_ability(stats['active2'])

    stats['statBoost1'] = parse_stat(stats['statBoost1'])
    stats['statBoost2'] = parse_stat(stats['statBoost2'])
    stats['statsUnknown1'] = stats.get(stats['statsUnknown1'], None)  # parse_stat(stat_genes['statsUnknown1'])
    stats['statsUnknown2'] = stats.get(stats['statsUnknown2'], None)  # parse_stat(stat_genes['statsUnknown2'])

    stats['element'] = parse_element(stats['element'])

    return stats


def parse_visual_genes(genes):
    visual_genes = {}
    visual_genes['raw'] = genes

    visual_raw_kai = "".join(__genesToKai(visual_genes['raw']).split(' '))

    for ki in range(0, len(visual_raw_kai)):
        stat_trait = visual_traits.get(int(ki / 4), None)
        kai = visual_raw_kai[ki]
        value_num = __kai2dec(kai)
        visual_genes[stat_trait] = value_num

    visual_genes['gender'] = 'male' if visual_genes['gender'] == 1 else 'female'
    return visual_genes


def hero2str(hero):

    if isinstance(hero['info']['class'], int):
        c = parse_class(hero['info']['class'])
        sc = parse_class(hero['info']['subClass'])
        r = parse_rarity(hero['info']['rarity'])
        l = hero['state']['level']
    else:
        c = hero['info']['class']
        sc = hero['info']['subClass']
        r = hero['info']['rarity']
        l = hero['state']['level']

    return str(hero['id']) + " " + r.title() + " " + c.title() + "/" + sc.title() + " lvl " + str(l)


def __genesToKai(genes):
    BASE = len(ALPHABET)

    buf = ''
    while genes >= BASE:
        mod = int(genes % BASE)
        buf = ALPHABET[int(mod)] + buf
        genes = (genes - mod) // BASE

    # Add the last 4 (finally).
    buf = ALPHABET[int(genes)] + buf

    # Pad with leading 1s.
    buf = buf.rjust(48, '1')
    buf = buf[0:48]

    return ' '.join(buf[i:i + 4] for i in range(0, len(buf), 4))


def __kai2dec(kai):
    return ALPHABET.index(kai)


def parse_names(names_raw_string):
    names_raw_string = names_raw_string\
        .replace("\\xf3", "ó") \
        .replace("\\xf2", "ò") \
        .replace("\\xe9", "é") \
        .replace("\\xe1", "á") \
        .replace("\\xc9", "É") \
        .replace("\\xed", "í") \
        .replace("\\xfa", "ú") \
        .replace("\\xec", "ì")

    if "\\x" in names_raw_string:
        raise Exception("Unhandled unicode found")

    return json.loads(names_raw_string)



from . import auction_core


class Auction:

    def __init__(self, auction_address, rpc_address, logger):
        self.auction_address = auction_address
        self.rpc_address = rpc_address
        self.logger = logger

    def bid(self, token_id, bid_amount_wei, private_key, nonce, gas_price_gwei, tx_timeout_seconds):
        return auction_core.bid(self.auction_address, token_id, bid_amount_wei, private_key, nonce, gas_price_gwei, tx_timeout_seconds, self.rpc_address, self.logger)

    def create_auction(self, token_id, starting_price_wei, ending_price_wei, duration, winner, private_key, nonce, gas_price_gwei, tx_timeout_seconds):
        return auction_core.create_auction(self.auction_address, token_id, starting_price_wei, ending_price_wei, duration, winner, private_key, nonce, gas_price_gwei,
                                           tx_timeout_seconds, self.rpc_address, self.logger)

    def cancel_auction(self, token_id, private_key, nonce, gas_price_gwei, tx_timeout_seconds):
        return auction_core.cancel_auction(self.auction_address, token_id, private_key, nonce, gas_price_gwei, tx_timeout_seconds, self.rpc_address, self.logger)

    def is_on_auction(self, token_id):
        return auction_core.is_on_auction(self.auction_address, token_id, self.rpc_address)

    def get_auction(self, token_id):
        return auction_core.get_auction(self.auction_address, token_id, self.rpc_address)

    def get_auctions(self, token_ids):
        return auction_core.get_auctions(self.auction_address, token_ids, self.rpc_address)

    def get_user_auctions(self, user):
        return auction_core.get_user_auctions(self.auction_address, user, self.rpc_address)

    def total_auctions(self):
        return auction_core.total_auctions(self.auction_address, self.rpc_address)

    def auctions(self, index):
        return auction_core.auctions(self.auction_address, index, self.rpc_address)



type = {
    'solo': 1,
    'squad': 3,
    'war': 9
}

background = {
    'desert': 0,
    'forest': 2,
    'plains': 4,
    'island': 6,
    'swamp': 8,
    'mountains': 10,
    'city': 12,
    'arctic': 14
}

stat = {
    'strength': 0,
    'agility': 2,
    'intelligence': 4,
    'wisdom': 6,
    'luck': 8,
    'vitality': 10,
    'endurance': 12,
    'dexterity': 14
}


def string2id(attr, label):
    if attr == "type":
        return type.get(label, None)
    elif attr == 'background':
        return background.get(label, None)
    elif attr == 'stat':
        return stat.get(label, None)

    return None


def id2string(attr, label):
    if attr == "type":
        return [k for k, v in type.items() if v == label][0]
    elif attr == 'background':
        return [k for k, v in background.items() if v == label][0]
    elif attr == 'stat':
        return [k for k, v in stat.items() if v == label][0]

    return None



import logging
import sys
import time
import summoning.assisting as assisting
from web3 import Web3

SD = 'serendale'
CV = 'crystalvale'

if __name__ == "__main__":
    log_format = '%(asctime)s|%(name)s|%(levelname)s: %(message)s'

    logger = logging.getLogger("DFK-assisting")
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.INFO, format=log_format, stream=sys.stdout)

    rpc_server = 'https://subnets.avax.network/defi-kingdoms/dfk-chain/rpc'
    hero_id = 420

    ret = assisting.is_on_rent(assisting.CRYSTALVALE_CONTRACT_ADDRESS, 404, rpc_server)
    logger.info("Hero "+str(hero_id) + (" is up for rent" if ret else " is not available to rent"))


FAIL_ON_NOT_FOUND = False

types = {
    1: "attemptBased",
    #2: "timeBased",
    #3: "well",
}


def parse_type(id):
    value = types.get(id, None)
    if FAIL_ON_NOT_FOUND and value is None:
        raise Exception("Quest type not found")
    return value


def human_readable_quest(raw_quest):
    if raw_quest is None:
        return None

    quest = {}
    i = 0

    if len(raw_quest) == 10:  # v2
        quest['id'] = raw_quest[i]
        i = i + 1
        quest['address'] = raw_quest[i]
        i = i + 1
        quest['level'] = raw_quest[i]
        i = i + 1
        quest['heroes'] = raw_quest[i]
        i = i + 1
        quest['player'] = raw_quest[i]
        i = i + 1
        quest['startBlock'] = raw_quest[i]
        i = i + 1
        quest['startTime'] = raw_quest[i]
        i = i + 1
        quest['completeAtTime'] = raw_quest[i]
        i = i + 1
        quest['attempts'] = raw_quest[i]
        i = i + 1
        quest['type'] = parse_type(raw_quest[i])
    else:  # TODO:// v1 - remove once old quests have been migrated
        quest['id'] = raw_quest[i]
        i = i + 1
        quest['address'] = raw_quest[i]
        i = i + 1
        quest['heroes'] = raw_quest[i]
        i = i + 1
        quest['player'] = raw_quest[i]
        i = i + 1
        quest['startTime'] = raw_quest[i]
        i = i + 1
        quest['startBlock'] = raw_quest[i]
        i = i + 1
        quest['completeAtTime'] = raw_quest[i]
        i = i + 1
        quest['attempts'] = raw_quest[i]
        i = i + 1
        quest['type'] = parse_type(raw_quest[i])

    return quest


def human_readable_quest_results(quest_results, very_human=False):
    quest_rewards = {}
    rewards = quest_results['reward']
    xps = quest_results['xp']
    skill_ups = quest_results['skillUp']

    for rew in rewards:
        hero_id = rew['args']['heroId']
        item = rew['args']['rewardItem'] if 'rewardItem' in rew['args'] else rew['args']['reward']
        qty = rew['args']['itemQuantity'] if 'itemQuantity' in rew['args'] else rew['args']['amount']

        if very_human:
            if item.upper() == '0x72Cb10C6bfA5624dD07Ef608027E366bd690048F'.upper()\
                    or item.upper() == '0x4f60a160D8C2DDdaAfe16FCC57566dB84D674BD6'.upper():
                item = "JEWEL"
                qty = qty / 1e18
            elif item.upper() == '0x3a4EDcf3312f44EF027acfd8c21382a5259936e7'.upper()\
                    or item.upper() == '0x576C260513204392F0eC0bc865450872025CB1cA'.upper():
                item = "DFKGOLD"
                qty = qty / 1e3

        if hero_id not in quest_rewards:
            quest_rewards[hero_id] = {
                "rewards": {},
                "xpEarned": 0.0,
                "skillUp": 0.0
            }

        if item not in quest_rewards[hero_id]["rewards"]:
            quest_rewards[hero_id]["rewards"][item] = 0

        quest_rewards[hero_id]["rewards"][item] += qty

    for xp in xps:
        hero_id = xp['args']['heroId']
        xp_earned = xp['args']['xpEarned']

        if hero_id not in quest_rewards:
            quest_rewards[hero_id] = {
                "rewards": {},
                "xpEarned": 0.0,
                "skillUp": 0.0
            }

        quest_rewards[hero_id]["xpEarned"] += xp_earned

    for su in skill_ups:
        hero_id = su['args']['heroId']
        su_earned = su['args']['skillUp']

        if hero_id not in quest_rewards:
            quest_rewards[hero_id] = {
                "rewards": {},
                "xpEarned": 0.0,
                "skillUp": 0.0
            }

        quest_rewards[hero_id]["skillUp"] += su_earned

    return quest_rewards


FAIL_ON_NOT_FOUND = False

egg_types = {
    0: "blue",
    1: "grey"
}

egg_rarity = {
    0: "common",
    1: "uncommon",
    2: "rare",
    3: "legendary",
    4: "mythic",
}


egg_elements = {
    0: 'fire',
    1: 'water',
    2: 'earth',
    3: 'wind',
    4: 'lightning',
    5: 'ice',
    6: 'light',
    7: 'dark',
}


def parse_egg_type(egg_type):
    value = egg_types.get(egg_type, None)
    if FAIL_ON_NOT_FOUND and value is None:
        raise Exception("Egg type not found")
    return value


def parse_egg_rarity(rarity):
    value = egg_rarity.get(rarity, None)
    if FAIL_ON_NOT_FOUND and value is None:
        raise Exception("Egg rarity not found")
    return value


def parse_egg_element(element):
    value = egg_elements.get(element, None)
    if FAIL_ON_NOT_FOUND and value is None:
        raise Exception("Egg element not found")
    return value


def human_readable_pet(raw_pet):
    if raw_pet is None:
        return None

    pet = {}
    i = 0

    pet['id'] = raw_pet[i]
    i = i + 1
    pet['originId'] = raw_pet[i]
    i = i + 1
    pet['name'] = raw_pet[i]
    i = i + 1
    pet['season'] = raw_pet[i]
    i = i + 1
    pet['eggType'] = parse_egg_type(raw_pet[i])
    i = i + 1
    pet['rarity'] = parse_egg_rarity(raw_pet[i])
    i = i + 1
    pet['element'] = parse_egg_element(raw_pet[i])
    i = i + 1
    pet['bonusCount'] = raw_pet[i]
    i = i + 1
    pet['profBonus'] = raw_pet[i]
    i = i + 1
    pet['profBonusScalar'] = raw_pet[i]
    i = i + 1
    pet['craftBonus'] = raw_pet[i]
    i = i + 1
    pet['combatBonusScalar'] = raw_pet[i]
    i = i + 1
    pet['appearance'] = raw_pet[i]
    i = i + 1
    pet['background'] = raw_pet[i]
    i = i + 1
    pet['shiny'] = raw_pet[i]
    i = i + 1
    pet['hungryAt'] = raw_pet[i]
    i = i + 1
    pet['equippableAt   '] = raw_pet[i]
    i = i + 1
    pet['equippedTo'] = raw_pet[i]
    i = i + 1




    return pet


import web3.contract
from web3 import Web3

SERENDALE_CONTRACT_ADDRESS = '0x573e407Be90a50EAbA28748cbb62Ff9d6038A3e9'
SERENDALE_CHAIN_ID = 1666600000
CRYSTALVALE_CONTRACT_ADDRESS = '0x739B1666c2956f601f095298132773074c3E184b'
CRYSTALVALE_CHAIN_ID = 53935

ABI = '''
    [
        {"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"previousAdmin","type":"address"},{"indexed":false,"internalType":"address","name":"newAdmin","type":"address"}],"name":"AdminChanged","type":"event"},
        {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"implementation","type":"address"}],"name":"Upgraded","type":"event"},{"stateMutability":"payable","type":"fallback"},
        {"inputs":[],"name":"admin","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"address","name":"newAdmin","type":"address"}],"name":"changeAdmin","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[],"name":"implementation","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"address","name":"newImplementation","type":"address"}],"name":"upgradeTo","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"address","name":"newImplementation","type":"address"},{"internalType":"bytes","name":"data","type":"bytes"}],"name":"upgradeToAndCall","outputs":[],"stateMutability":"payable","type":"function"},{"stateMutability":"payable","type":"receive"},
        {"anonymous":false,"inputs":[{"indexed":true,"internalType":"uint256","name":"heroId","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"arrivalChainId","type":"uint256"}],"name":"HeroArrived","type":"event"},
        {"anonymous":false,"inputs":[{"indexed":true,"internalType":"uint256","name":"heroId","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"arrivalChainId","type":"uint256"}],"name":"HeroSent","type":"event"},
        {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"previousOwner","type":"address"},{"indexed":true,"internalType":"address","name":"newOwner","type":"address"}],"name":"OwnershipTransferred","type":"event"},
        {"anonymous":false,"inputs":[{"indexed":false,"internalType":"uint256","name":"_srcChainId","type":"uint256"},{"indexed":false,"internalType":"bytes32","name":"_srcAddress","type":"bytes32"}],"name":"SetTrustedRemote","type":"event"},
        {"inputs":[],"name":"assistingAuction","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},
        {"inputs":[{"internalType":"bytes32","name":"_srcAddress","type":"bytes32"},{"internalType":"uint256","name":"_srcChainId","type":"uint256"},{"internalType":"bytes","name":"_message","type":"bytes"},{"internalType":"address","name":"_executor","type":"address"}],"name":"executeMessage","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"uint256","name":"_chainId","type":"uint256"}],"name":"getTrustedRemote","outputs":[{"internalType":"bytes32","name":"trustedRemote","type":"bytes32"}],"stateMutability":"view","type":"function"},
        {"inputs":[],"name":"heroes","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},
        {"inputs":[{"internalType":"address","name":"_messageBus","type":"address"},{"internalType":"address","name":"_heroes","type":"address"},{"internalType":"address","name":"_assistingAuction","type":"address"}],"name":"initialize","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[],"name":"messageBus","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},
        {"inputs":[],"name":"msgGasLimit","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
        {"inputs":[],"name":"owner","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},
        {"inputs":[],"name":"renounceOwnership","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"uint256","name":"_heroId","type":"uint256"},{"internalType":"uint256","name":"_dstChainId","type":"uint256"}],"name":"sendHero","outputs":[],"stateMutability":"payable","type":"function"},
        {"inputs":[{"internalType":"address","name":"_assistingAuction","type":"address"}],"name":"setAssistingAuctionAddress","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"address","name":"_messageBus","type":"address"}],"name":"setMessageBus","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"uint256","name":"_msgGasLimit","type":"uint256"}],"name":"setMsgGasLimit","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"uint256","name":"_srcChainId","type":"uint256"},{"internalType":"bytes32","name":"_srcAddress","type":"bytes32"}],"name":"setTrustedRemote","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"address","name":"newOwner","type":"address"}],"name":"transferOwnership","outputs":[],"stateMutability":"nonpayable","type":"function"},
        {"inputs":[{"internalType":"address","name":"initialLogic","type":"address"},{"internalType":"address","name":"initialAdmin","type":"address"},{"internalType":"bytes","name":"_data","type":"bytes"}],"stateMutability":"payable","type":"constructor"}
    ]
    '''


def block_explorer_link(contract_address, txid):
    if hasattr(contract_address, 'address'):
        contract_address = str(contract_address.address)
    contract_address = str(contract_address).upper()
    if contract_address == SERENDALE_CONTRACT_ADDRESS.upper():
        return 'https://explorer.harmony.one/tx/' + str(txid)
    elif contract_address == CRYSTALVALE_CONTRACT_ADDRESS.upper():
        return 'https://subnets.avax.network/defi-kingdoms/dfk-chain/explorer/tx/' + str(txid)
    else:
        return str(txid)


def assisting_auction(realm_contract, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract = Web3.toChecksumAddress(realm_contract)
    contract = w3.eth.contract(contract, abi=ABI)
    result = contract.functions.assistingAuction().call()

    return result


def heroes(realm_contract, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    contract = Web3.toChecksumAddress(realm_contract)
    contract = w3.eth.contract(contract, abi=ABI)
    result = contract.functions.heroes().call()

    return result


def send_hero(origin_realm_contract_address, hero_id, destination_chain_id, bridge_fee_in_wei, private_key, nonce, gas_price_gwei, tx_timeout_seconds, rpc_address, logger):

    w3 = Web3(Web3.HTTPProvider(rpc_address))
    account = w3.eth.account.privateKeyToAccount(private_key)
    w3.eth.default_account = account.address

    origin_realm_contract_address = Web3.toChecksumAddress(origin_realm_contract_address)
    contract = w3.eth.contract(origin_realm_contract_address, abi=ABI)

    tx = contract.functions.sendHero(hero_id, destination_chain_id)

    if isinstance(gas_price_gwei, dict):   # dynamic fee
        tx = tx.buildTransaction(
            {'maxFeePerGas': w3.toWei(gas_price_gwei['maxFeePerGas'], 'gwei'),
             'maxPriorityFeePerGas': w3.toWei(gas_price_gwei['maxPriorityFeePerGas'], 'gwei'),
             'value': bridge_fee_in_wei, 'nonce': nonce})
    else:   # legacy
        tx = tx.buildTransaction({'gasPrice': w3.toWei(gas_price_gwei, 'gwei'), 'value': bridge_fee_in_wei, 'nonce': nonce})

    logger.debug("Signing transaction")
    signed_tx = w3.eth.account.sign_transaction(tx, private_key=private_key)
    logger.debug("Sending transaction " + str(tx))
    ret = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    logger.debug("Transaction successfully sent !")
    logger.info(
        "Waiting for transaction " + block_explorer_link(origin_realm_contract_address, signed_tx.hash.hex()) + " to be mined")
    tx_receipt = w3.eth.wait_for_transaction_receipt(transaction_hash=signed_tx.hash, timeout=tx_timeout_seconds,
                                                     poll_latency=2)
    logger.info("Transaction mined !")

    return tx_receipt


from web3 import Web3
from .erc1155 import ITEMS as ERC1155_ITEMS

ITEMS = ERC1155_ITEMS

ABI = """
        [
            {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"operator","type":"address"},{"indexed":false,"internalType":"bool","name":"approved","type":"bool"}],"name":"ApprovalForAll","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"account","type":"address"}],"name":"Paused","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"previousAdminRole","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"newAdminRole","type":"bytes32"}],"name":"RoleAdminChanged","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleGranted","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleRevoked","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"operator","type":"address"},{"indexed":true,"internalType":"address","name":"from","type":"address"},{"indexed":true,"internalType":"address","name":"to","type":"address"},{"indexed":false,"internalType":"uint256[]","name":"ids","type":"uint256[]"},{"indexed":false,"internalType":"uint256[]","name":"values","type":"uint256[]"}],"name":"TransferBatch","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"operator","type":"address"},{"indexed":true,"internalType":"address","name":"from","type":"address"},{"indexed":true,"internalType":"address","name":"to","type":"address"},{"indexed":false,"internalType":"uint256","name":"id","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"value","type":"uint256"}],"name":"TransferSingle","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":false,"internalType":"string","name":"value","type":"string"},{"indexed":true,"internalType":"uint256","name":"id","type":"uint256"}],"name":"URI","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"account","type":"address"}],"name":"Unpaused","type":"event"},
            {"inputs":[],"name":"DEFAULT_ADMIN_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"MINTER_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"PAUSER_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"address","name":"account","type":"address"},{"internalType":"uint256","name":"id","type":"uint256"}],"name":"balanceOf","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"address[]","name":"accounts","type":"address[]"},{"internalType":"uint256[]","name":"ids","type":"uint256[]"}],"name":"balanceOfBatch","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"address","name":"account","type":"address"},{"internalType":"uint256","name":"id","type":"uint256"},{"internalType":"uint256","name":"value","type":"uint256"}],"name":"burn","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"address","name":"account","type":"address"},{"internalType":"uint256[]","name":"ids","type":"uint256[]"},{"internalType":"uint256[]","name":"values","type":"uint256[]"}],"name":"burnBatch","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[],"name":"decimals","outputs":[{"internalType":"uint8","name":"","type":"uint8"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"}],"name":"getRoleAdmin","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"uint256","name":"index","type":"uint256"}],"name":"getRoleMember","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"}],"name":"getRoleMemberCount","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"grantRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"hasRole","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"initialize","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"string","name":"uri","type":"string"}],"name":"initialize","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"address","name":"account","type":"address"},{"internalType":"address","name":"operator","type":"address"}],"name":"isApprovedForAll","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"address","name":"to","type":"address"},{"internalType":"uint256","name":"id","type":"uint256"},{"internalType":"uint256","name":"amount","type":"uint256"},{"internalType":"bytes","name":"data","type":"bytes"}],"name":"mint","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"address","name":"to","type":"address"},{"internalType":"uint256[]","name":"ids","type":"uint256[]"},{"internalType":"uint256[]","name":"amounts","type":"uint256[]"},{"internalType":"bytes","name":"data","type":"bytes"}],"name":"mintBatch","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[],"name":"name","outputs":[{"internalType":"string","name":"","type":"string"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"pause","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[],"name":"paused","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"renounceRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"revokeRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"address","name":"from","type":"address"},{"internalType":"address","name":"to","type":"address"},{"internalType":"uint256[]","name":"ids","type":"uint256[]"},{"internalType":"uint256[]","name":"amounts","type":"uint256[]"},{"internalType":"bytes","name":"data","type":"bytes"}],"name":"safeBatchTransferFrom","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"address","name":"from","type":"address"},{"internalType":"address","name":"to","type":"address"},{"internalType":"uint256","name":"id","type":"uint256"},{"internalType":"uint256","name":"amount","type":"uint256"},{"internalType":"bytes","name":"data","type":"bytes"}],"name":"safeTransferFrom","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"address","name":"operator","type":"address"},{"internalType":"bool","name":"approved","type":"bool"}],"name":"setApprovalForAll","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"bytes4","name":"interfaceId","type":"bytes4"}],"name":"supportsInterface","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"symbol","outputs":[{"internalType":"string","name":"","type":"string"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"unpause","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"","type":"uint256"}],"name":"uri","outputs":[{"internalType":"string","name":"","type":"string"}],"stateMutability":"view","type":"function"}
        ]
        """


def address2item(address):
    address = address.upper().strip()
    for item in ITEMS:
        if item[0].upper() == address:
            return item
    return None


def address2symbol(address):
    address = address.upper().strip()
    for item in ITEMS:
        if item[0].upper() == address:
            return item[1]
    return None


def all_items():
    return ITEMS.copy()


def block_explorer_link(txid):
    return 'https://explorer.harmony.one/tx/' + str(txid)


def balance_of(token_address, account_address, token_id, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))
    contract_address = Web3.toChecksumAddress(token_address)
    contract = w3.eth.contract(contract_address, abi=ABI)
    result = contract.functions.balanceOf(account_address, token_id).call()

    return result


def decimals(token_address, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))
    contract_address = Web3.toChecksumAddress(token_address)
    contract = w3.eth.contract(contract_address, abi=ABI)
    result = contract.functions.decimals().call()

    return result


def symbol(token_address, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))
    contract_address = Web3.toChecksumAddress(token_address)
    contract = w3.eth.contract(contract_address, abi=ABI)
    result = contract.functions.symbol().call()

    return result


def uri(token_address, token_id, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))
    contract_address = Web3.toChecksumAddress(token_address)
    contract = w3.eth.contract(contract_address, abi=ABI)
    result = contract.functions.uri(token_id).call()

    return result


def mint(token_address, receiver_address, token_id, amount, data, private_key, nonce, gas_price_gwei, tx_timeout_seconds, rpc_address, logger):
    w3 = Web3(Web3.HTTPProvider(rpc_address))
    account = w3.eth.account.privateKeyToAccount(private_key)
    w3.eth.default_account = account.address

    contract_address = Web3.toChecksumAddress(token_address)
    contract = w3.eth.contract(contract_address, abi=ABI)
    tx = contract.functions.mint(receiver_address, token_id, amount, data).buildTransaction(
        {'gasPrice': w3.toWei(gas_price_gwei, 'gwei'), 'nonce': nonce})
    logger.debug("Signing transaction")
    signed_tx = w3.eth.account.sign_transaction(tx, private_key=private_key)
    logger.debug("Sending transaction " + str(tx))
    ret = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    logger.debug("Transaction successfully sent !")
    logger.info("Waiting for transaction " + block_explorer_link(signed_tx.hash.hex()) + " to be mined")
    tx_receipt = w3.eth.wait_for_transaction_receipt(transaction_hash=signed_tx.hash, timeout=tx_timeout_seconds,
                                                     poll_latency=2)
    logger.info("Transaction mined !")
    return tx_receipt




import logging
import sys
from web3 import Web3
import hero.utils.utils as hero_utils
import auctions.hero.sale_auctions as hero_sales
import auctions.hero.rent_auctions as hero_rental
from auctions.auction import Auction
import auctions.land as land_auctions
import auctions.pet as pet_auctions
import auctions.utils.utils as auction_utils


if __name__ == "__main__":
    log_format = '%(asctime)s|%(name)s|%(levelname)s: %(message)s'

    logger = logging.getLogger("DFK-auctions")
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.INFO, format=log_format, stream=sys.stdout)

    rpc_server = 'https://api.harmony.one'
    logger.info("Using RPC server " + rpc_server)

    graphql = 'https://defi-kingdoms-community-api-gateway-co06z8vi.uc.gateway.dev/graphql'

    auctions = hero_sales.get_open_auctions(graphql, 0, 10)
    logger.info("Hero sale auctions:")
    for auction in auctions:
        logger.info(str(auction))

    logger.info("\n")

    logger.info("Hero rental auctions:")
    auctions = hero_rental.get_open_auctions(graphql, 0, 10)
    for auction in auctions:
        logger.info(str(auction))

    #w3 = Web3(Web3.HTTPProvider(rpc_server))
    #private_key = ""  # set private key
    #account_address = w3.eth.account.privateKeyToAccount(private_key).address

    # serendale hero auction
    #logger.info(hero_sales.get_auction(hero_sales.SERENDALE_CONTRACT_ADDRESS, 181373, 'https://api.harmony.one'))
    # hero_sales.bid_hero(hero_sales.SERENDALE_CONTRACT_ADDRESS, 181373, hero_sales.ether2wei(100), private_key, 'w3.eth.getTransactionCount(account_address), 50, 30)

    # crystalvale hero auction
    #cv_hero_auctions = Auction(hero_sales.CRYSTALVALE_CONTRACT_ADDRESS, 'https://subnets.avax.network/defi-kingdoms/dfk-chain/rpc', logger)
    #logger.info(auction_utils.human_readable_auction(cv_hero_auctions.get_auction(hero_utils.sd2cv_cv_hero_id(250))))
    # cv_hero_auctions.bid_hero(hero_utils.sd2cv_cv_hero_id(250), hero_sales.ether2wei(100), private_key, w3.eth.getTransactionCount(account_address), 50, 30)

    logger.info("\n")

    # lands
    user = '0x2E7669F61eA77F02445A015FBdcFe2DE47083E02'

    land_auction = Auction(land_auctions.AUCTION_CONTRACT_ADDRESS, rpc_server, logger)
    logger.info("Total land auctions: " + str(land_auction.total_auctions()))
    logger.info("Auctions of " + user + ": " + str(land_auction.get_user_auctions(user)))

    for i in range(0, min(10, land_auction.total_auctions())):
        auction = auction_utils.human_readable_auction(land_auction.auctions(i))
        logger.info("Auction index " + str(i) + ": " + str(auction))

    if land_auction.is_on_auction(17):
        auction = land_auction.get_auction(17)  # must be an open auction
        auction = auction_utils.human_readable_auction(auction)
        logger.info(str(auction))

    logger.info("\n")

    # pets
    pet_auction = Auction(pet_auctions.AUCTION_CONTRACT_ADDRESS, rpc_server, logger)
    logger.info("Total pet auctions: " + str(pet_auction.total_auctions()))
    logger.info("Auctions of " + user + ": " + str(pet_auction.get_user_auctions(user)))

    for i in range(0, min(10, pet_auction.total_auctions())):
        auction = auction_utils.human_readable_auction(pet_auction.auctions(i))
        logger.info("Auction index " + str(i) + ": " + str(auction))

    if pet_auction.is_on_auction(17):
        auction = pet_auction.get_auction(17)  # must be an open auction
        auction = auction_utils.human_readable_auction(auction)
        logger.info(str(auction))





import logging
from web3 import Web3
import sys
import time
import meditation.meditation as meditation

if __name__ == "__main__":
    log_format = '%(asctime)s|%(name)s|%(levelname)s: %(message)s'

    logger = logging.getLogger("DFK-meditation")
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.INFO, format=log_format, stream=sys.stdout)

    # NOTE: make sure you use the realm where your hero currently is
    realm = 'crystalvale'
    if realm == 'serendale':
        rpc_server = 'https://api.harmony.one'
        contract_address = meditation.SERENDALE_CONTRACT_ADDRESS
        gas_price_gwei = 115
        tx_timeout_seconds = 30        
    elif realm == 'crystalvale':
        rpc_server = 'https://subnets.avax.network/defi-kingdoms/dfk-chain/rpc'
        contract_address = meditation.CRYSTALVALE_CONTRACT_ADDRESS
        gas_price_gwei = {'maxFeePerGas': 2, 'maxPriorityFeePerGas': 2}
        tx_timeout_seconds = 30

    logger.info("Using RPC server " + rpc_server)

    # Customize your hero info, stats, & credentials here
    hero_id = 1
    level = 1    
    private_key = None  # set private key
    account_address = '0x...'
    # Stats to update should be full words, lowercase. 
    # See meditation.stat2id() for examples
    stat1 = 'strength'
    stat2 = 'endurance'
    stat3 = 'luck'
    attunement_crystal_address = meditation.ZERO_ADDRESS

    w3 = Web3(Web3.HTTPProvider(rpc_server))
    # NOTE that you will have to have approved Shvas runes & the meditation contract 
    # before calling these functions, and that account address has all the needed
    # runes
    active_meditations = meditation.get_active_meditations(contract_address, account_address, rpc_server)
    logger.info("Pending meditation on address " + str(account_address) + ": "+str(active_meditations))

    
    required_runes = meditation.get_required_runes(contract_address, level, rpc_server)
    meditation.start_meditation(contract_address, hero_id, stat1, stat2, stat3,
                                attunement_crystal_address, private_key, w3.eth.getTransactionCount(account_address),
                                gas_price_gwei, tx_timeout_seconds, rpc_server, logger)
    hero_meditation = meditation.get_hero_meditation(contract_address, hero_id, rpc_server)
    logger.info("Pending meditation "+str(hero_meditation))
    logger.info("Waiting 20 seconds to complete meditation")
    time.sleep(20)
    tx_receipt = meditation.complete_meditation(contract_address, hero_id, private_key, w3.eth.getTransactionCount(account_address),
                                   gas_price_gwei, tx_timeout_seconds, rpc_server, logger)

    level_up = meditation.parse_meditation_results(contract_address, tx_receipt, rpc_server)
                
    meditation_results_msg = ""
    for hero_id in level_up:
        inc = level_up[hero_id]
        meditation_results_msg += f"""
            ```
            Hero: {hero_id}
                Level: {inc['new level']}  STAM: +{inc.get("STAMINA", {}).get("increase", 0)}  HP: +{inc.get("HP",{}).get("increase", 0)}
                STR: +{inc.get("STR", {}).get("increase", 0)}  DEX: +{inc.get("DEX", {}).get("increase", 0)}
                AGI: +{inc.get("AGI", {}).get("increase", 0)}  VIT: +{inc.get("VIT", {}).get("increase", 0)}
                END: +{inc.get("END", {}).get("increase", 0)}  INT: +{inc.get("INT", {}).get("increase", 0)}
                WIS: +{inc.get("WIS", {}).get("increase", 0)}  LCK: +{inc.get("LCK", {}).get("increase", 0)}
            ```"""

    msg = f"Completed level up: {meditation_results_msg}"
    logger.info(msg)




def human_readable_auction(auction):
    human_readable = {}
    human_readable['seller'] = auction[0]
    human_readable['tokenId'] = auction[1]
    human_readable['startingPrice'] = auction[2]
    human_readable['endingPrice'] = auction[3]
    human_readable['duration'] = auction[4]
    human_readable['startedAt'] = auction[5]
    human_readable['winner'] = auction[6]
    human_readable['open'] = auction[7]

    return human_readable


import logging
import sys
import quests.wishing_well as wishing_well

if __name__ == "__main__":
    log_format = '%(asctime)s|%(name)s|%(levelname)s: %(message)s'

    logger = logging.getLogger("DFK-wishing well quest")
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.INFO, format=log_format, stream=sys.stdout)

    rpc_server = 'https://api.harmony.one'
    logger.info("Using RPC server " + rpc_server)

    level = wishing_well.quest_level(rpc_server)
    logger.info("Quest level "+str(level))

    hero_id = 1  # <your hero id here>
    stamina = wishing_well.get_current_stamina(hero_id, rpc_server)
    logger.info("Current stamina on hero " + str(hero_id) + ": " + str(stamina))

    #w3 = Web3(Web3.HTTPProvider(rpc_server))
    #gas_price_gwei = 10
    #private_key = # set private key
    #account_address = w3.eth.account.privateKeyToAccount(private_key).address
    #wishing_well.start_quest(hero_id, 5, private_key, w3.eth.getTransactionCount(account_address), gas_price_gwei, 30, rpc_server, logger)
    #time.sleep(60)
    #tx_receipt = wishing_well.complete_quest(hero_id, private_key, w3.eth.getTransactionCount(account_address), gas_price_gwei, 30, rpc_server, logger)

    #quest_result = wishing_well.parse_complete_quest_receipt(tx_receipt, rpc_server)
    #logger.info("Quest earned " + str(quest_result['tear']) + " tears and " + str(quest_result['xp']) + " xp")






from . import hero_core


class Hero:
    def __init__(self, contract_address, rpc_address, logger=None):
        self.contract_address = contract_address
        self.rpc_address = rpc_address
        self.logger = logger

    def transfer(self, hero_id, owner_private_key, owner_nonce, receiver_address, gas_price_gwei, tx_timeout_seconds):
        return hero_core.transfer(self.contract_address, hero_id, owner_private_key, owner_nonce, receiver_address, gas_price_gwei, tx_timeout_seconds, self.rpc_address, self.logger)

    def get_owner(self, hero_id, block_identifier="latest"):
        return hero_core.get_owner(self.contract_address, hero_id, self.rpc_address, block_identifier)

    def get_users_heroes(self, user_address, block_identifier="latest"):
        return hero_core.get_users_heroes(self.contract_address, user_address, self.rpc_address, block_identifier)

    def is_approved_for_all(self, owner, operator):
        return hero_core.is_approved_for_all(self.contract_address, owner, operator, self.rpc_address)

    def get_hero(self, hero_id, block_identifier="latest"):
        return hero_core.get_hero(self.contract_address, hero_id, self.rpc_address, block_identifier)

    @staticmethod
    def human_readable_hero(raw_hero, hero_male_first_names=None, hero_female_first_names=None, hero_last_names=None):
        return hero_core.human_readable_hero(raw_hero, hero_male_first_names, hero_female_first_names, hero_last_names)



import logging
import sys
from web3 import Web3
import consumable.consumable as consumable

if __name__ == "__main__":
    log_format = '%(asctime)s|%(name)s|%(levelname)s: %(message)s'

    logger = logging.getLogger("DFK-consumable")
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.INFO, format=log_format, stream=sys.stdout)

    rpc_server = 'https://api.harmony.one'
    logger.info("Using RPC server " + rpc_server)

    stamina_vial_address = '0x959ba19508827d1ed2333B1b503Bd5ab006C710e'

    w3 = Web3(Web3.HTTPProvider(rpc_server))
    private_key = None  # set private key
    account_address = w3.eth.account.privateKeyToAccount(private_key).address
    gas_price_gwei = 40
    tx_timeout_seconds = 30

    hero_id = 1
    consumable.consume_item(consumable.SERENDALE_CONTRACT_ADDRESS, stamina_vial_address, hero_id, private_key, w3.eth.getTransactionCount(account_address), gas_price_gwei, tx_timeout_seconds, rpc_server, logger)


import requests
from web3 import Web3

SERENDALE_CONTRACT_ADDRESS = '0x13a65B9F8039E2c032Bc022171Dc05B30c3f2892'
CRYSTALVALE_CONTRACT_ADDRESS = '0xc390fAA4C7f66E4D62E59C231D5beD32Ff77BEf0'


ABI = """
        [
            {"inputs":[{"internalType":"address","name":"_heroCoreAddress","type":"address"},{"internalType":"address","name":"_geneScienceAddress","type":"address"},{"internalType":"address","name":"_jewelTokenAddress","type":"address"},{"internalType":"address","name":"_gaiaTearsAddress","type":"address"},{"internalType":"address","name":"_statScienceAddress","type":"address"},{"internalType":"uint256","name":"_cut","type":"uint256"}],"stateMutability":"nonpayable","type":"constructor"},
            {"anonymous":false,"inputs":[{"indexed":false,"internalType":"uint256","name":"auctionId","type":"uint256"},{"indexed":true,"internalType":"uint256","name":"tokenId","type":"uint256"}],"name":"AuctionCancelled","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":false,"internalType":"uint256","name":"auctionId","type":"uint256"},{"indexed":true,"internalType":"address","name":"owner","type":"address"},{"indexed":true,"internalType":"uint256","name":"tokenId","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"startingPrice","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"endingPrice","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"duration","type":"uint256"},{"indexed":false,"internalType":"address","name":"winner","type":"address"}],"name":"AuctionCreated","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":false,"internalType":"uint256","name":"auctionId","type":"uint256"},{"indexed":true,"internalType":"uint256","name":"tokenId","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"totalPrice","type":"uint256"},{"indexed":false,"internalType":"address","name":"winner","type":"address"}],"name":"AuctionSuccessful","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"owner","type":"address"},{"indexed":false,"internalType":"uint256","name":"crystalId","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"heroId","type":"uint256"}],"name":"CrystalOpen","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":false,"internalType":"uint256","name":"crystalId","type":"uint256"},{"indexed":true,"internalType":"address","name":"owner","type":"address"},{"indexed":false,"internalType":"uint256","name":"summonerId","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"assistantId","type":"uint256"},{"indexed":false,"internalType":"uint16","name":"generation","type":"uint16"},{"indexed":false,"internalType":"uint256","name":"createdBlock","type":"uint256"},{"indexed":false,"internalType":"uint8","name":"summonerTears","type":"uint8"},{"indexed":false,"internalType":"uint8","name":"assistantTears","type":"uint8"},{"indexed":false,"internalType":"address","name":"bonusItem","type":"address"}],"name":"CrystalSummoned","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"previousOwner","type":"address"},{"indexed":true,"internalType":"address","name":"newOwner","type":"address"}],"name":"OwnershipTransferred","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"account","type":"address"}],"name":"Paused","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"previousAdminRole","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"newAdminRole","type":"bytes32"}],"name":"RoleAdminChanged","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleGranted","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleRevoked","type":"event"},
            {"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"account","type":"address"}],"name":"Unpaused","type":"event"},
            {"inputs":[],"name":"DEFAULT_ADMIN_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"MODERATOR_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"auctionHeroCore","outputs":[{"internalType":"contract IHeroCore","name":"","type":"address"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"baseCooldown","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"baseSummonFee","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_tokenId","type":"uint256"},{"internalType":"uint256","name":"_bidAmount","type":"uint256"}],"name":"bid","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"components":[{"internalType":"uint256","name":"id","type":"uint256"},{"components":[{"internalType":"uint256","name":"summonedTime","type":"uint256"},{"internalType":"uint256","name":"nextSummonTime","type":"uint256"},{"internalType":"uint256","name":"summonerId","type":"uint256"},{"internalType":"uint256","name":"assistantId","type":"uint256"},{"internalType":"uint32","name":"summons","type":"uint32"},{"internalType":"uint32","name":"maxSummons","type":"uint32"}],"internalType":"struct IHeroTypes.SummoningInfo","name":"summoningInfo","type":"tuple"},{"components":[{"internalType":"uint256","name":"statGenes","type":"uint256"},{"internalType":"uint256","name":"visualGenes","type":"uint256"},{"internalType":"enum IHeroTypes.Rarity","name":"rarity","type":"uint8"},{"internalType":"bool","name":"shiny","type":"bool"},{"internalType":"uint16","name":"generation","type":"uint16"},{"internalType":"uint32","name":"firstName","type":"uint32"},{"internalType":"uint32","name":"lastName","type":"uint32"},{"internalType":"uint8","name":"shinyStyle","type":"uint8"},{"internalType":"uint8","name":"class","type":"uint8"},{"internalType":"uint8","name":"subClass","type":"uint8"}],"internalType":"struct IHeroTypes.HeroInfo","name":"info","type":"tuple"},{"components":[{"internalType":"uint256","name":"staminaFullAt","type":"uint256"},{"internalType":"uint256","name":"hpFullAt","type":"uint256"},{"internalType":"uint256","name":"mpFullAt","type":"uint256"},{"internalType":"uint16","name":"level","type":"uint16"},{"internalType":"uint64","name":"xp","type":"uint64"},{"internalType":"address","name":"currentQuest","type":"address"},{"internalType":"uint8","name":"sp","type":"uint8"},{"internalType":"enum IHeroTypes.HeroStatus","name":"status","type":"uint8"}],"internalType":"struct IHeroTypes.HeroState","name":"state","type":"tuple"},{"components":[{"internalType":"uint16","name":"strength","type":"uint16"},{"internalType":"uint16","name":"intelligence","type":"uint16"},{"internalType":"uint16","name":"wisdom","type":"uint16"},{"internalType":"uint16","name":"luck","type":"uint16"},{"internalType":"uint16","name":"agility","type":"uint16"},{"internalType":"uint16","name":"vitality","type":"uint16"},{"internalType":"uint16","name":"endurance","type":"uint16"},{"internalType":"uint16","name":"dexterity","type":"uint16"},{"internalType":"uint16","name":"hp","type":"uint16"},{"internalType":"uint16","name":"mp","type":"uint16"},{"internalType":"uint16","name":"stamina","type":"uint16"}],"internalType":"struct IHeroTypes.HeroStats","name":"stats","type":"tuple"},{"components":[{"internalType":"uint16","name":"strength","type":"uint16"},{"internalType":"uint16","name":"intelligence","type":"uint16"},{"internalType":"uint16","name":"wisdom","type":"uint16"},{"internalType":"uint16","name":"luck","type":"uint16"},{"internalType":"uint16","name":"agility","type":"uint16"},{"internalType":"uint16","name":"vitality","type":"uint16"},{"internalType":"uint16","name":"endurance","type":"uint16"},{"internalType":"uint16","name":"dexterity","type":"uint16"},{"internalType":"uint16","name":"hpSm","type":"uint16"},{"internalType":"uint16","name":"hpRg","type":"uint16"},{"internalType":"uint16","name":"hpLg","type":"uint16"},{"internalType":"uint16","name":"mpSm","type":"uint16"},{"internalType":"uint16","name":"mpRg","type":"uint16"},{"internalType":"uint16","name":"mpLg","type":"uint16"}],"internalType":"struct IHeroTypes.HeroStatGrowth","name":"primaryStatGrowth","type":"tuple"},{"components":[{"internalType":"uint16","name":"strength","type":"uint16"},{"internalType":"uint16","name":"intelligence","type":"uint16"},{"internalType":"uint16","name":"wisdom","type":"uint16"},{"internalType":"uint16","name":"luck","type":"uint16"},{"internalType":"uint16","name":"agility","type":"uint16"},{"internalType":"uint16","name":"vitality","type":"uint16"},{"internalType":"uint16","name":"endurance","type":"uint16"},{"internalType":"uint16","name":"dexterity","type":"uint16"},{"internalType":"uint16","name":"hpSm","type":"uint16"},{"internalType":"uint16","name":"hpRg","type":"uint16"},{"internalType":"uint16","name":"hpLg","type":"uint16"},{"internalType":"uint16","name":"mpSm","type":"uint16"},{"internalType":"uint16","name":"mpRg","type":"uint16"},{"internalType":"uint16","name":"mpLg","type":"uint16"}],"internalType":"struct IHeroTypes.HeroStatGrowth","name":"secondaryStatGrowth","type":"tuple"},{"components":[{"internalType":"uint16","name":"mining","type":"uint16"},{"internalType":"uint16","name":"gardening","type":"uint16"},{"internalType":"uint16","name":"foraging","type":"uint16"},{"internalType":"uint16","name":"fishing","type":"uint16"}],"internalType":"struct IHeroTypes.HeroProfessions","name":"professions","type":"tuple"}],"internalType":"struct IHeroTypes.Hero","name":"_hero","type":"tuple"}],"name":"calculateSummoningCost","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_tokenId","type":"uint256"}],"name":"cancelAuction","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_tokenId","type":"uint256"}],"name":"cancelAuctionWhenPaused","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[],"name":"cooldownPerGen","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_tokenId","type":"uint256"},{"internalType":"uint128","name":"_startingPrice","type":"uint128"},{"internalType":"uint128","name":"_endingPrice","type":"uint128"},{"internalType":"uint64","name":"_duration","type":"uint64"},{"internalType":"address","name":"_winner","type":"address"}],"name":"createAuction","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"","type":"uint256"}],"name":"crystals","outputs":[{"internalType":"address","name":"owner","type":"address"},{"internalType":"uint256","name":"summonerId","type":"uint256"},{"internalType":"uint256","name":"assistantId","type":"uint256"},{"internalType":"uint16","name":"generation","type":"uint16"},{"internalType":"uint256","name":"createdBlock","type":"uint256"},{"internalType":"uint256","name":"heroId","type":"uint256"},{"internalType":"uint8","name":"summonerTears","type":"uint8"},{"internalType":"uint8","name":"assistantTears","type":"uint8"},{"internalType":"address","name":"bonusItem","type":"address"},{"internalType":"uint32","name":"maxSummons","type":"uint32"},{"internalType":"uint32","name":"firstName","type":"uint32"},{"internalType":"uint32","name":"lastName","type":"uint32"},{"internalType":"uint8","name":"shinyStyle","type":"uint8"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_rarityRoll","type":"uint256"},{"internalType":"uint256","name":"_rarityMod","type":"uint256"}],"name":"determineRarity","outputs":[{"internalType":"enum IHeroTypes.Rarity","name":"","type":"uint8"}],"stateMutability":"pure","type":"function"},
            {"inputs":[],"name":"enabled","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"randomNumber","type":"uint256"},{"internalType":"uint256","name":"digits","type":"uint256"},{"internalType":"uint256","name":"offset","type":"uint256"}],"name":"extractNumber","outputs":[{"internalType":"uint256","name":"result","type":"uint256"}],"stateMutability":"pure","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_tokenId","type":"uint256"}],"name":"getAuction","outputs":[{"internalType":"uint256","name":"auctionId","type":"uint256"},{"internalType":"address","name":"seller","type":"address"},{"internalType":"uint256","name":"startingPrice","type":"uint256"},{"internalType":"uint256","name":"endingPrice","type":"uint256"},{"internalType":"uint256","name":"duration","type":"uint256"},{"internalType":"uint256","name":"startedAt","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_crystalId","type":"uint256"}],"name":"getCrystal","outputs":[{"components":[{"internalType":"address","name":"owner","type":"address"},{"internalType":"uint256","name":"summonerId","type":"uint256"},{"internalType":"uint256","name":"assistantId","type":"uint256"},{"internalType":"uint16","name":"generation","type":"uint16"},{"internalType":"uint256","name":"createdBlock","type":"uint256"},{"internalType":"uint256","name":"heroId","type":"uint256"},{"internalType":"uint8","name":"summonerTears","type":"uint8"},{"internalType":"uint8","name":"assistantTears","type":"uint8"},{"internalType":"address","name":"bonusItem","type":"address"},{"internalType":"uint32","name":"maxSummons","type":"uint32"},{"internalType":"uint32","name":"firstName","type":"uint32"},{"internalType":"uint32","name":"lastName","type":"uint32"},{"internalType":"uint8","name":"shinyStyle","type":"uint8"}],"internalType":"struct ICrystalTypes.HeroCrystal","name":"","type":"tuple"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_tokenId","type":"uint256"}],"name":"getCurrentPrice","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"}],"name":"getRoleAdmin","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"address","name":"_address","type":"address"}],"name":"getUserAuctions","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"address","name":"_address","type":"address"}],"name":"getUserCrystals","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"grantRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"hasRole","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"increasePerGen","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"increasePerSummon","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_tokenId","type":"uint256"}],"name":"isOnAuction","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[],"name":"jewelToken","outputs":[{"internalType":"contract IJewelToken","name":"","type":"address"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"newSummonCooldown","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_crystalId","type":"uint256"}],"name":"open","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[],"name":"owner","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"ownerCut","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"paused","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_crystalId","type":"uint256"}],"name":"rechargeCrystal","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[],"name":"renounceOwnership","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"renounceRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"revokeRole","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"address[]","name":"_feeAddresses","type":"address[]"},{"internalType":"uint256[]","name":"_feePercents","type":"uint256[]"}],"name":"setFees","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"address","name":"_geneScienceAddress","type":"address"}],"name":"setGeneScienceAddress","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"address","name":"_statScienceAddress","type":"address"}],"name":"setStatScienceAddress","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_newSummonCooldown","type":"uint256"},{"internalType":"uint256","name":"_baseCooldown","type":"uint256"},{"internalType":"uint256","name":"_cooldownPerGen","type":"uint256"}],"name":"setSummonCooldowns","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_baseSummonFee","type":"uint256"},{"internalType":"uint256","name":"_increasePerSummon","type":"uint256"},{"internalType":"uint256","name":"_increasePerGen","type":"uint256"}],"name":"setSummonFees","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[],"name":"statScience","outputs":[{"internalType":"contract IStatScience","name":"","type":"address"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"_summonerId","type":"uint256"},{"internalType":"uint256","name":"_assistantId","type":"uint256"},{"internalType":"uint16","name":"_summonerTears","type":"uint16"},{"internalType":"uint16","name":"_assistantTears","type":"uint16"},{"internalType":"address","name":"_bonusItem","type":"address"}],"name":"summonCrystal","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"bytes4","name":"interfaceId","type":"bytes4"}],"name":"supportsInterface","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},
            {"inputs":[],"name":"toggleEnabled","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"address","name":"newOwner","type":"address"}],"name":"transferOwnership","outputs":[],"stateMutability":"nonpayable","type":"function"},
            {"inputs":[{"internalType":"address","name":"","type":"address"},{"internalType":"uint256","name":"","type":"uint256"}],"name":"userAuctions","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"address","name":"","type":"address"},{"internalType":"uint256","name":"","type":"uint256"}],"name":"userCrystals","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},
            {"inputs":[{"internalType":"uint256","name":"blockNumber","type":"uint256"}],"name":"vrf","outputs":[{"internalType":"bytes32","name":"result","type":"bytes32"}],"stateMutability":"view","type":"function"}
        ]
        """

AUCTIONS_OPEN_GRAPHQL_QUERY = """
                        query {
                          saleAuctions(skip: %d, first: %d, orderBy: startedAt, orderDirection: desc, where: {open: true, winner: null}) {
                            id
                            seller {
                                name
                            }
                            tokenId {
                              id
                              owner {
                                owner
                              }
                              statGenes
                              generation
                              rarity
                              mainClass
                              subClass
                              strength
                              intelligence
                              wisdom
                              luck
                              agility
                              vitality
                              endurance
                              dexterity
                              level
                              summons
                              maxSummons
                              summonerId {
                                id
                              }
                              assistantId {
                                id
                              }
                            }
                            startingPrice
                            endingPrice
                            startedAt
                            duration
                            winner {
                              id
                              name
                            }
                            open
                            
                          }
                          
                        }
                        """

AUCTIONS_TOKEN_IDS_GRAPHQL_QUERY = """
                        query {
                          saleAuctions(orderBy: startedAt, orderDirection: desc, where: {open: true, tokenId_in: %s }) {
                            id
                            seller {
                                name
                            }
                            tokenId {
                              id
                              owner {
                                owner
                              }
                              
                              statGenes
                              generation
                              rarity
                              mainClass
                              subClass
                              summons
                              maxSummons
                              summonerId {
                                id
                              }
                              assistantId {
                                id
                              }
                            }
                            startingPrice
                            endingPrice
                            startedAt
                            duration
                            winner {
                              id
                              name
                            }
                            open

                          }

                        }
                        """


def block_explorer_link(txid):
    return 'https://explorer.harmony.one/tx/' + str(txid)


def bid_hero(contract_address, token_id, bid_amount_wei, private_key, nonce, gas_price_gwei, tx_timeout_seconds, rpc_address, logger=None):
    w3 = Web3(Web3.HTTPProvider(rpc_address))
    account = w3.eth.account.privateKeyToAccount(private_key)
    w3.eth.default_account = account.address

    sales_auction_contract_address = Web3.toChecksumAddress(contract_address)
    sales_auction_contract = w3.eth.contract(sales_auction_contract_address, abi=ABI)

    if logger is not None:
        logger.info("Biding " + str(wei2ether(bid_amount_wei)) + " on hero id " + str(token_id))
    tx = sales_auction_contract.functions.bid(token_id, bid_amount_wei).buildTransaction(
        {'gasPrice': w3.toWei(gas_price_gwei, 'gwei'), 'nonce': nonce})

    if logger is not None:
        logger.info("Signing transaction")
    signed_tx = w3.eth.account.sign_transaction(tx, private_key=private_key)
    if logger is not None:
        logger.info("Sending transaction " + str(tx))
    ret = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    if logger is not None:
        logger.info("Transaction successfully sent !")
        logger.info("Waiting for transaction " + block_explorer_link(signed_tx.hash.hex()) + " to be mined")
    tx_receipt = w3.eth.wait_for_transaction_receipt(transaction_hash=signed_tx.hash, timeout=tx_timeout_seconds, poll_latency=2)

    if logger is not None:
        logger.info("Transaction mined !")
        logger.info(str(tx_receipt))

    return tx_receipt


def create_auction(contract_address, token_id, starting_price_wei, ending_price_wei, duration, winner, private_key, nonce, gas_price_gwei, tx_timeout_seconds, rpc_address, logger=None):
    w3 = Web3(Web3.HTTPProvider(rpc_address))
    account = w3.eth.account.privateKeyToAccount(private_key)
    w3.eth.default_account = account.address

    sales_auction_contract_address = Web3.toChecksumAddress(contract_address)
    sales_auction_contract = w3.eth.contract(sales_auction_contract_address, abi=ABI)

    if logger is not None:
        logger.info("Auctioning " + str(token_id) + " (starting price=" + str(wei2ether(starting_price_wei)) + ", ending price=" + str(wei2ether(ending_price_wei)) + ", duration=" + str(duration) + ", private sale buyer=" + str(winner) + ")")
    tx = sales_auction_contract.functions.createAuction(token_id, starting_price_wei, ending_price_wei, duration, winner).buildTransaction(
        {'gasPrice': w3.toWei(gas_price_gwei, 'gwei'), 'nonce': nonce})
    if logger is not None:
        logger.info("Signing transaction")
    signed_tx = w3.eth.account.sign_transaction(tx, private_key=private_key)
    if logger is not None:
        logger.info("Sending transaction " + str(tx))
    ret = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    if logger is not None:
        logger.info("Transaction successfully sent !")
        logger.info("Waiting for transaction " + block_explorer_link(signed_tx.hash.hex()) + " to be mined")
    tx_receipt = w3.eth.wait_for_transaction_receipt(transaction_hash=signed_tx.hash, timeout=tx_timeout_seconds, poll_latency=2)
    if logger is not None:
        logger.info("Transaction mined !")
        logger.info(str(tx_receipt))
    return tx_receipt


def cancel_auction(contract_address, token_id, private_key, nonce, gas_price_gwei, tx_timeout_seconds, rpc_address, logger=None):
    w3 = Web3(Web3.HTTPProvider(rpc_address))
    account = w3.eth.account.privateKeyToAccount(private_key)
    w3.eth.default_account = account.address

    sales_auction_contract_address = Web3.toChecksumAddress(contract_address)
    sales_auction_contract = w3.eth.contract(sales_auction_contract_address, abi=ABI)

    tx = sales_auction_contract.functions.cancelAuction(token_id).buildTransaction(
        {'gasPrice': w3.toWei(gas_price_gwei, 'gwei'), 'nonce': nonce})
    if logger is not None:
        logger.info("Signing transaction")
    signed_tx = w3.eth.account.sign_transaction(tx, private_key=private_key)
    if logger is not None:
        logger.info("Sending transaction " + str(tx))
    ret = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    if logger is not None:
        logger.info("Transaction successfully sent !")
        logger.info("Waiting for transaction " + block_explorer_link(signed_tx.hash.hex()) + " to be mined")
    tx_receipt = w3.eth.wait_for_transaction_receipt(transaction_hash=signed_tx.hash, timeout=tx_timeout_seconds, poll_latency=2)
    if logger is not None:
        logger.info("Transaction mined !")
        logger.info(str(tx_receipt))
    return tx_receipt


def is_on_auction(contract_address, token_id, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    sales_auction_contract_address = Web3.toChecksumAddress(contract_address)
    sales_auction_contract = w3.eth.contract(sales_auction_contract_address, abi=ABI)
    return sales_auction_contract.functions.isOnAuction(token_id).call()


def get_auction(contract_address, token_id, rpc_address):
    w3 = Web3(Web3.HTTPProvider(rpc_address))

    sales_auction_contract_address = Web3.toChecksumAddress(contract_address)
    sales_auction_contract = w3.eth.contract(sales_auction_contract_address, abi=ABI)
    result = sales_auction_contract.functions.getAuction(token_id).call()
    auction = {}
    auction['id'] = result[0]
    auction['owner'] = result[1]
    auction['startingPrice'] = result[2]
    auction['endingPrice'] = result[3]
    auction['duration'] = result[4]
    auction['startedAt'] = result[5]

    return auction


def get_open_auctions(graphql_address, skip=0, count=1000):

    r = requests.post(graphql_address, json={'query': AUCTIONS_OPEN_GRAPHQL_QUERY % (skip, count)})

    if r.status_code != 200:
        raise Exception("HTTP error " + str(r.status_code) + ": " + r.text)
    data = r.json()
    return data['data']['saleAuctions']


def get_hero_open_auctions(graphql_address, hero_ids):
    str_hero_ids = "["
    for id in hero_ids:
        str_hero_ids = str_hero_ids + "\"" + str(id) + "\", "
    str_hero_ids = str_hero_ids + "]"

    r = requests.post(graphql_address, json={'query': AUCTIONS_TOKEN_IDS_GRAPHQL_QUERY % str_hero_ids})
    if r.status_code != 200:
        raise Exception("HTTP error " + str(r.status_code) + ": " + r.text)
    data = r.json()
    return data['data']['saleAuctions']


def auction2hero(auction):
    ah = auction['tokenId']

    hero = {}
    hero['id'] = ah['id']
    hero['info'] = {}
    hero['info']['class'] = ah['mainClass'].lower()
    hero['info']['subClass'] = ah['subClass'].lower()
    hero['info']['rarity'] = ah['rarity']
    hero['info']['level'] = ah['level']
    hero['info']['statGenes'] = ah['statGenes']
    hero['info']['generation'] = ah['generation']

    hero['stats'] = {}
    hero['stats']['strength'] = ah['strength']
    hero['stats']['agility'] = ah['agility']
    hero['stats']['intelligence'] = ah['intelligence']
    hero['stats']['wisdom'] = ah['wisdom']
    hero['stats']['luck'] = ah['luck']
    hero['stats']['vitality'] = ah['vitality']
    hero['stats']['endurance'] = ah['endurance']
    hero['stats']['dexterity'] = ah['dexterity']

    hero['summoningInfo'] = {}
    hero['summoningInfo']['summonerId'] = ah['summonerId']
    hero['summoningInfo']['assistantId'] = ah['assistantId']
    hero['summoningInfo']['maxSummons'] = ah['maxSummons']
    hero['summoningInfo']['summons'] = ah['summons']

    return hero


def wei2ether(wei):
    return float(wei) / 1000000000000000000


def ether2wei(ether):
    return int(ether * 1000000000000000000)




SERENDALE_QUEST_CONTRACT_ADDRESS_V1 = '0xe4154B6E5D240507F9699C730a496790A722DF19'
CRYSTALVALE_QUEST_CONTRACT_ADDRESSES_V2 = [
    (0, 'wJEWEL-xJEWEL', '0xd3d8ff8e42C2eD51FabE4BA34080C6ac79395f24'),
    (1, 'CRYSTAL-AVAX', '0x8eDA0ceA7a90E794B33708Cc0768727A1A612f3d'),
    (2, 'CRYSTAL-wJEWEL', '0xC4839Fb9A5466878168EaE3fD58c647B71475b61'),
    (3, 'CRYSTAL-USDC', '0x6FEF23498877bC4c3940ebE121dd7D138BdA4e11'),
    (4, 'ETH-USDC', '0xdeF7cBeE7d0B62037616ee26BCAc1C8364f53476'),
    (5, 'wJEWEL-USDC', '0xaac3933Faa3B668304C9276d10CA88853463BD42'),
    (6, 'CRYSTAL-ETH', '0x810e1fF51fDd58c474c66A31013713D1A17BF458'),
    (7, 'CRYSTAL-BTC.b', '0x706916dbC3b66d89632708CC193080ea05E0534A'),
    (8, 'CRYSTAL-KLAY', '0x1fCc67a01525fd715A67bCcbF73665Fb3dBE76c7'),
    (9, 'JEWEL-KLAY', '0x2A70aA48f9dBF859239ae5E7f98fe95aE27A6CD4'),
    (10, 'JEWEL-AVAX', '0xA0d17554F09047d65E0ae0e76CD8923A9525183c'),
    (11, 'JEWEL-BTC.b', '0x3391B9384AC66C7Aa3BF4A75A4f441942B1dCf30'),
    (12, 'JEWEL-ETH', '0xbaEc39Dd81b964B57bc5fa5f5421Cd82185409E6'),
    (13, 'BTC.b-USDC', '0x045838dBfb8026520E872c8298F4Ed542B81Eaca')
]


def get_crystalvale_contract_info(address):
    address = str(address).upper()
    for addr in CRYSTALVALE_QUEST_CONTRACT_ADDRESSES_V2:
        if addr[2].upper() == address:
            return addr
    return None


def get_cv_pool_id_contract_address(pool_id):
    for addr in CRYSTALVALE_QUEST_CONTRACT_ADDRESSES_V2:
        if addr[0] == pool_id:
            return addr[2]
    return None


def get_cv_liquidity_pair_contract_address(liquidity_pair):
    for addr in CRYSTALVALE_QUEST_CONTRACT_ADDRESSES_V2:
        if addr[1] == liquidity_pair:
            return addr[2]
    return None



import logging
import sys
import raffle.raffle_master as raffle_master
import raffle.utils.utils as raffle_utils
from web3 import Web3

if __name__ == "__main__":
    log_format = '%(asctime)s|%(name)s|%(levelname)s: %(message)s'

    logger = logging.getLogger("DFK-raffle")
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.INFO, format=log_format, stream=sys.stdout)

    rpc_server = 'https://api.harmony.one'
    logger.info("Using RPC server " + rpc_server)

    w3 = Web3(Web3.HTTPProvider(rpc_server))
    private_key = ""  # set private key
    account_address = "0x"  # w3.eth.account.privateKeyToAccount(private_key).address
    gas_price_gwei = 115
    tx_timeout_seconds = 30

    # Get current raffles
    current_raffle = raffle_master.get_current_raffle_data(rpc_server)

    raffle_log = "Current raffle:"
    for r in current_raffle[0]:
        raffle_log = raffle_log + "\n\t" + str(raffle_utils.human_readable_raffle(r))
    logger.info(raffle_log)

    raffle_type_log = "Current raffle type:"
    for r in current_raffle[1]:
        raffle_type_log = raffle_type_log + "\n\t" + str(raffle_utils.human_readable_raffle_type(r))
    logger.info(raffle_type_log)

    # Get raffle by id
    raffle_by_id = raffle_master.get_raffle_list([1, 2], rpc_server)
    raffle_by_id_log = "Getting raffle by ids:"
    for r in raffle_by_id:
        raffle_by_id_log = raffle_by_id_log + "\n\t" + str(raffle_utils.human_readable_raffle(r))
    logger.info(raffle_by_id_log)

    # Get raffle allowed maximum entries
    raffle_id = [current_raffle[0][0][0]]
    raffle_max_entries = raffle_master.get_raffle_tickets_allowance_list(raffle_id, rpc_server)
    logger.info("Maximum allowed entries for raffle {} : {}".format(raffle_id[0],  raffle_max_entries[0]))

    # Enter raffle
    raffle_master.enter_raffle(3555, 1,
                               private_key, w3.eth.getTransactionCount(account_address),
                               gas_price_gwei, tx_timeout_seconds, rpc_server, logger)



import logging
import sys
from web3 import Web3
import pets.pet_core as pet_core
import pets.hatchery as hatchery
import pets.exchange as exchange
import pets.utils.utils as pet_utils

if __name__ == "__main__":
    log_format = '%(asctime)s|%(name)s|%(levelname)s: %(message)s'

    logger = logging.getLogger("DFK-pet")
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.INFO, format=log_format, stream=sys.stdout)

    rpc_server = 'https://subnets.avax.network/defi-kingdoms/dfk-chain/rpc'
    logger.info("Using RPC server " + rpc_server)

    user = '0x2E7669F61eA77F02445A015FBdcFe2DE47083E02'

    # pet NFT
    logger.info("Total existing pet: {}".format(str(pet_core.next_pet_id(pet_core.CRYSTALVALE_CONTRACT_ADDRESS, rpc_server))))
    user_balance = pet_core.balance_of(pet_core.CRYSTALVALE_CONTRACT_ADDRESS, user, rpc_server)
    logger.info("Player {} is owner of {} pets".format(user, user_balance))

    user_pets = pet_core.get_user_pets(pet_core.CRYSTALVALE_CONTRACT_ADDRESS, user, rpc_server)
    for pet in user_pets:
        logger.info(str(pet_utils.human_readable_pet(pet)))

    # exchange
    logger.info("Total pet exchange: {}".format(exchange.total_exchanges(exchange.CRYSTALVALE_CONTRACT_ADDRESS, rpc_server)))
    logger.info("Player {} has {} pet exchange pending".format(user, len(
        exchange.get_user_pet_exchanges(exchange.CRYSTALVALE_CONTRACT_ADDRESS, user, rpc_server))))

    # hatching
    incubating_eggs = hatchery.get_user_eggs(hatchery.CRYSTALVALE_CONTRACT_ADDRESS, user, rpc_server)
    logger.info("Player {} is incubating {} eggs".format(user, len(incubating_eggs)))

    blue_egg_cost = hatchery.egg_type_costs(hatchery.CRYSTALVALE_CONTRACT_ADDRESS, 0, rpc_server)
    grey_egg_cost = hatchery.egg_type_costs(hatchery.CRYSTALVALE_CONTRACT_ADDRESS, 1, rpc_server)
    logger.info("Incubating cost of blue egg: {}".format(str(blue_egg_cost)))
    logger.info("Incubating cost of grey egg: {}".format(str(grey_egg_cost)))

    private_key = None  # set private key
    gas_price_gwei = 115
    tx_timeout = 30
    w3 = Web3(Web3.HTTPProvider(rpc_server))
    account_address = w3.eth.account.privateKeyToAccount(private_key).address

    hatchery.incubate_egg(hatchery.CRYSTALVALE_CONTRACT_ADDRESS,0, 2, private_key, w3.eth.getTransactionCount(account_address), gas_price_gwei, tx_timeout, rpc_server, logger)
    hatchery.crack(hatchery.CRYSTALVALE_CONTRACT_ADDRESS, 404, private_key, w3.eth.getTransactionCount(account_address), gas_price_gwei, tx_timeout,
                  rpc_server, logger)


import logging
import sys
import time
from web3 import Web3
import dex.erc20 as erc20
import stone_carver.stone_carver as stone_carver

if __name__ == "__main__":
    log_format = '%(asctime)s|%(name)s|%(levelname)s: %(message)s'

    logger = logging.getLogger("DFK-stone-carver")
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.INFO, format=log_format, stream=sys.stdout)

    rpc_server = 'https://api.harmony.one'
    logger.info("Using RPC server " + rpc_server)

    lesser_chaos_stone = erc20.symbol2item('DFKLCHSST', 'serendale')
    lesser_chaos_stone_address = erc20.symbol2address('DFKLCHSST', 'serendale')

    # Request recipe for one Stone
    stone = stone_carver.get_recipe(lesser_chaos_stone_address, stone_carver.SERENDALE_CONTRACT_ADDRESS, rpc_server)
    logger.info(f"{lesser_chaos_stone[2]}: {stone}")

    # Request Stone Carver availability
    availability = stone_carver.get_availability(stone_carver.SERENDALE_CONTRACT_ADDRESS, rpc_server)
    logger.info(f"Working until: {availability[0]} / Reopens at: {availability[1]}")

    # Displays Working until Timestamp / whether SC is currently working
    working = stone_carver.working_until(stone_carver.SERENDALE_CONTRACT_ADDRESS, rpc_server)
    logger.info(f"Working until: {working} / Currently Working: {True if time.time() < working else False}")

    # Crafting Lesser Chaos Stone
    w3 = Web3(Web3.HTTPProvider(rpc_server))
    private_key = None  # set private key
    account_address = w3.eth.account.privateKeyToAccount(private_key).address
    gas_price_gwei = 100
    tx_timeout_seconds = 30
    stone_carver.carve_stone(stone_carver.SERENDALE_CONTRACT_ADDRESS, lesser_chaos_stone_address, 1, private_key, w3.eth.getTransactionCount(account_address), gas_price_gwei, tx_timeout_seconds, rpc_server, logger)



from . import quest_core_v1


class Quest:
    def __init__(self, rpc_address, logger):
        self.rpc_address = rpc_address
        self.logger = logger

    def start_quest(self, quest_address, hero_ids, attempts, private_key, nonce, gas_price_gwei, tx_timeout_seconds):
        return quest_core_v1.start_quest(quest_address, hero_ids, attempts, private_key, nonce, gas_price_gwei,
                                         tx_timeout_seconds, self.rpc_address, self.logger)

    def start_quest_with_data(self, quest_address, data, hero_ids, attempts, private_key, nonce, gas_price_gwei,
                              tx_timeout_seconds):
        return quest_core_v1.start_quest_with_data(quest_address, data, hero_ids, attempts, private_key, nonce,
                                                   gas_price_gwei, tx_timeout_seconds, self.rpc_address, self.logger)

    def complete_quest(self, hero_id, private_key, nonce, gas_price_gwei, tx_timeout_seconds):
        return quest_core_v1.complete_quest(hero_id, private_key, nonce, gas_price_gwei, tx_timeout_seconds,
                                            self.rpc_address, self.logger)

    def parse_complete_quest_receipt(self, tx_receipt):
        return quest_core_v1.parse_complete_quest_receipt(tx_receipt, self.rpc_address)

    def cancel_quest(self, hero_id, private_key, nonce, gas_price_gwei, tx_timeout_seconds):
        return quest_core_v1.cancel_quest(hero_id, private_key, nonce, gas_price_gwei, tx_timeout_seconds,
                                          self.rpc_address, self.logger)

    def hero_to_quest_id(self, hero_id):
        return quest_core_v1.hero_to_quest_id(hero_id, self.rpc_address)

    def is_hero_questing(self, hero_id):
        return self.hero_to_quest_id(hero_id) > 0

    def get_active_quest(self, address):
        return quest_core_v1.get_active_quest(address, self.rpc_address)

    def get_hero_quest(self, hero_id):
        return quest_core_v1.get_hero_quest(hero_id, self.rpc_address)

    def get_quest(self, quest_id):
        return quest_core_v1.get_quest(quest_id, self.rpc_address)

    def get_quest_data(self, quest_id):
        return quest_core_v1.get_quest_data(quest_id, self.rpc_address)

    def quest_address_to_type(self, quest_address):
        return quest_core_v1.quest_address_to_type(quest_address, self.rpc_address)

    def get_current_stamina(self, hero_id):
        return quest_core_v1.get_current_stamina(hero_id, self.rpc_address)